
import copy
from typing import List, Optional, Set, Type, Union
import uuid

from pydantic import BaseModel

from authzee import exceptions
from authzee.grant import Grant
from authzee.grant_effect import GrantEffect
from authzee.grants_page import GrantsPage
from authzee.resource_action import ResourceAction
from authzee.resource_authz import ResourceAuthz


class StorageBackend:
    """Base class for ``Authzee`` storage. 

    Base classes must at least implement:

        - ``initialize`` - Initialize the storage backend. External connections should be created here
        - ``shutdown`` - Preemptively cleanup storage backend resources.
        - ``setup`` - One time setup for compute backend.
        - ``teardown`` - Remove resources created from ``setup()``.
        - ``add_grant`` - Add a grant to storage.
        - ``delete_grant`` - Delete a grant from storage.
        - ``get_grants_page`` - Retrieve a page of grants from storage. 
    
    Optionally ``async`` methods may also be created for calls to storage. 

        - ``add_grant_async``
        - ``delete_grant_async``
        - ``get_grants_page_async``

    The sub-class must also set the class vars:

        - ``async_enabled`` - The class has all ``async`` methods available.
        - ``process_safe`` - The storage is safe between processes ie. external to the running program.

    No error checking should be needed for validation of resources, resource_types etc. That should all be handled by ``Authzee``.

    Storage backends should store all arguments to the ``__init__`` method in ``self.kwargs``, 
    and all arguments to the ``initialize`` method in ``self.initialize_kwargs``.  
    These should be available if the compute backend needs to instantiate more instances of the storage backend.
    """

    async_enabled: bool = False
    process_safe: bool = False


    def __init__(self, *, default_page_size: int, **kwargs):
        self.default_page_size = default_page_size
        self.kwargs = kwargs
        self.kwargs['default_page_size'] = default_page_size
        self.initialize_kwargs = {}


    def initialize(
        self, 
        identity_types: Set[Type[BaseModel]],
        resource_authzs: List[ResourceAuthz]
    ) -> None:
        """Initialize the storage backend. 

        All creation of connections to external storage should take place here,
        but not creation of resources in the storage.  ie don't create tables.

        Parameters
        ----------
        identity_types : Set[Type[BaseModel]]
            Identity types that have been registered with ``Authzee``.
        resource_authzs : List[ResourceAuthz]
            ``ResourceAuthz`` instances that have been registered with ``Authzee``.
        """
        self.initialize_kwargs = {
            "identity_types": identity_types,
            "resource_authzs": resource_authzs
        }
        self._identity_types = identity_types
        self._resource_authzs = resource_authzs
    

    def shutdown(self) -> None:
        """Early clean up of storage backend resources.
        """
        pass


    def setup(self) -> None:
        """One time setup for storage backend resources.
        """
        pass

    
    def teardown(self) -> None:
        """Teardown and delete the results of ``setup()`` .
        """
        pass

    
    def add_grant(self, effect: GrantEffect, grant: Grant) -> Grant:
        """Add a grant. 

        Parameters
        ----------
        effect : GrantEffect
            The effect of the grant.
        grant : Grant
            The grant.

        Returns
        -------
        Grant
            The grant that has been added with additional information for the specific backend.

        Raises
        ------
        authzee.exceptions.MethodNotImplementedError
            Sub-classes must implement this method.
        """
        raise exceptions.MethodNotImplementedError()
    

    async def add_grant_async(self, effect: GrantEffect, grant: Grant) -> Grant:
        """Add a grant. 

        Parameters
        ----------
        effect : GrantEffect
            The effect of the grant.
        grant : Grant
            The grant.

        Returns
        -------
        Grant
            The grant that has been added with additional information for the specific backend.

        Raises
        ------
        authzee.exceptions.MethodNotImplementedError
            Sub-classes *may* implement this method if ``async`` is supported.
        """
        raise exceptions.MethodNotImplementedError()


    def delete_grant(self, effect: GrantEffect, uuid: str) -> None:
        """Delete a grant.

        Parameters
        ----------
        effect : GrantEffect
            The effect of the grant.
        uuid : str
            UUID of grant to delete.

        Raises
        ------
        authzee.exceptions.MethodNotImplementedError
            Sub-classes must implement this method.
        """
        raise exceptions.MethodNotImplementedError()


    async def delete_grant_async(self, effect: GrantEffect, uuid: str) -> None:
        """Delete a grant.

        Parameters
        ----------
        effect : GrantEffect
            The effect of the grant.
        uuid : str
            UUID of grant to delete.

        Raises
        ------
        authzee.exceptions.MethodNotImplementedError
            Sub-classes *may* implement this method if ``async`` is supported.
        """
        raise exceptions.MethodNotImplementedError()


    def get_grants_page(
        self,
        effect: GrantEffect,
        resource_type: Optional[Type[BaseModel]] = None,
        resource_action: Optional[ResourceAction] = None,
        page_size: Optional[int] = None,
        next_page_reference: Optional[BaseModel] = None
    ) -> GrantsPage:
        """Retrieve a page of grants matching the filters.

        If ``GrantsPage.next_page_reference`` is not ``None`` , there are more grants to retrieve.
        To get the next page, pass ``next_page_reference=GrantsPage.next_page_reference`` .

        **NOTE** - There is no guarantee of how many grants will be returned if any.

        Parameters
        ----------
        effect : GrantEffect
            The effect of the grant.
        resource_type : Optional[Type[BaseModel]], optional
            Filter by resource type.
            By default no filter is applied.
        resource_action : Optional[ResourceAction], optional
            Filter by `ResourceAction``. 
            By default no filter is applied.
        page_size : Optional[int], optional
            The suggested page size to return. 
            There is no guarantee of how much data will be returned if any.
            The default is set on the storage backend. 
        next_page_reference : Optional[BaseModel], optional
            The reference to the next page that is returned in ``GrantsPage``.
            By default this will return the 1st page.

        Returns
        -------
        GrantsPage
            The page of grants.

        Raises
        ------
        authzee.exceptions.MethodNotImplementedError
            Sub-classes must implement this method.
        """
        raise exceptions.MethodNotImplementedError()
    

    async def get_grants_page_async(
        self,
        effect: GrantEffect,
        resource_type: Optional[Type[BaseModel]] = None,
        resource_action: Optional[ResourceAction] = None,
        page_size: Optional[int] = None,
        next_page_reference: Optional[BaseModel] = None
    ) -> GrantsPage:
        """Retrieve a page of grants matching the filters.

        If ``GrantsPage.next_page_reference`` is not ``None`` , there are more grants to retrieve.
        To get the next page, pass ``next_page_reference=GrantsPage.next_page_reference`` .

        **NOTE** - There is no guarantee of how many grants will be returned if any.

        Parameters
        ----------
        effect : GrantEffect
            The effect of the grant.
        resource_type : Optional[Type[BaseModel]], optional
            Filter by resource type.
            By default no filter is applied.
        resource_action : Optional[ResourceAction], optional
            Filter by `ResourceAction``. 
            By default no filter is applied.
        page_size : Optional[int], optional
            The suggested page size to return. 
            There is no guarantee of how much data will be returned if any.
            The default is set on the storage backend. 
        next_page_reference : Optional[BaseModel], optional
            The reference to the next page that is returned in ``GrantsPage``.
            By default this will return the 1st page.

        Returns
        -------
        GrantsPage
            The page of grants.

        Raises
        ------
        authzee.exceptions.MethodNotImplementedError
            Sub-classes *may* implement this method if ``async`` is supported.
        """
        raise exceptions.MethodNotImplementedError()


    def _check_uuid(self, grant: Grant, generate_uuid: bool) -> Grant:
        """Check if a UUID is on a grant to add, optionally generate a UUID with UUID 4.


        Parameters
        ----------
        grant : Grant
            The grant to check.
        generate_uuid : bool
            Optionally generate a UUID and add it to the grant. 

        Returns
        -------
        Grant
            A deep copy of the passed in grant, optionally with a UUID 4 added.

        Raises
        ------
        authzee.exceptions.GrantUUIDError
            Grants that are being added should not have a UUID.
        """
        if grant.uuid is not None:
            raise exceptions.GrantUUIDError("Cannot create a grant that has a UUID.")

        grant = copy.deepcopy(grant)
        if generate_uuid == True:
            grant.uuid = str(uuid.uuid4())
        
        return grant
    

    def _real_page_size(self, page_size: Union[int, None]) -> int:
        if page_size is None:
            return self.default_page_size
    
        return page_size






import asyncio
from typing import Any, AsyncIterator, Callable, Coroutine, Iterator, Optional, Union

from pydantic import BaseModel

from authzee.grant import Grant
from authzee.grants_page import GrantsPage


class GrantIter: 

    def __init__(
        self, 
        next_page_callable: Callable[..., GrantsPage],
        page_size: Optional[int] = None,
        next_page_reference: Optional[BaseModel] = None,
        **next_page_kwargs: Any
    ):
        self._next_page_callable = next_page_callable
        self._page_size = page_size
        self._start_page_reference = next_page_reference
        self._next_page_reference = next_page_reference
        self._next_page_kwargs = next_page_kwargs

        self._grants_page: GrantsPage = None
        self._grants_iter: Iterator[Grant] = None


    def __iter__(self) -> Iterator:
        self._next_page_reference = self._start_page_reference
        self._grants_page = None

        return self


    def __next__(self) -> Grant:
        return self._get_next_item()
    

    def _get_next_item(self) -> Grant:
        if self._grants_page is None:
            self._grants_page = self._next_page_callable(
                page_size=self._page_size,
                next_page_reference=self._next_page_reference,
                **self._next_page_kwargs
            )
            self._grants_iter = iter(self._grants_page.grants)
            self._next_page_reference = self._grants_page.next_page_reference

        try:
            return next(self._grants_iter)
        except StopIteration:
            if self._next_page_reference is None:
                raise

            self._grants_page = None
            
            return self._get_next_item()
    

    def __aiter__(self) -> AsyncIterator:
        self._next_page_reference = self._start_page_reference
        self._grants_page = None

        return self


    async def __anext__(self) -> Grant:
        return await self._get_next_item_async()
    

    async def _get_next_item_async(self) -> Grant:
        if self._grants_page is None:
            self._grants_page = await self._next_page_callable(
                page_size=self._page_size,
                next_page_reference=self._next_page_reference,
                **self._next_page_kwargs
            )
            self._grants_iter = iter(self._grants_page.grants)
            self._next_page_reference = self._grants_page.next_page_reference

        try:
            return next(self._grants_iter)
        except StopIteration:
            if self._next_page_reference is None:
                raise StopAsyncIteration

            self._grants_page = None

            return await self._get_next_item_async()





import asyncio
import json
from typing import Any, Dict, List, Optional, Set, Type

from pydantic import BaseModel
from sqlalchemy import event, select
from sqlalchemy.ext.asyncio import async_sessionmaker, AsyncSession, create_async_engine


from authzee import exceptions
from authzee.grant import Grant
from authzee.grant_effect import GrantEffect
from authzee.grants_page import GrantsPage
from authzee.resource_action import ResourceAction
from authzee.resource_authz import ResourceAuthz
from authzee.storage.sql_storage_models import AllowGrantDB, Base, DenyGrantDB, ResourceActionDB, ResourceTypeDB
from authzee.storage.storage_backend import StorageBackend


class SQLNextPageRef(BaseModel):

    next_token: int


class SQLStorage(StorageBackend):

    async_enabled: bool = True
    process_safe: bool = True


    def __init__(
        self,
        *,
        sqlalchemy_async_engine_kwargs: Dict[str, Any],
        default_page_size: int = 1000
    ):
        super().__init__(
            default_page_size=default_page_size,
            sqlalchemy_async_engine_kwargs=sqlalchemy_async_engine_kwargs
        )
        self._sqlalchemy_async_engine_kwargs = sqlalchemy_async_engine_kwargs


    def initialize(
        self, 
        identity_types: Set[Type[BaseModel]],
        resource_authzs: List[ResourceAuthz]
    ) -> None:
        """Initialize the SQL storage backend. 

        Should only be called by the ``Authzee`` app.

        Parameters
        ----------
        identity_types : Set[Type[BaseModel]]
            Identity types that have been registered with ``Authzee``.
        resource_authzs : List[ResourceAuthz]
            ``ResourceAuthz`` instances that have been registered with ``Authzee``.
        """
        super().initialize(
            identity_types=identity_types,
            resource_authzs=resource_authzs
        )
        self._resource_type_lookup: Dict[str, Type[BaseModel]] = {
            authz.resource_type.__name__: authz.resource_type for authz in resource_authzs
        }
        self._resource_action_type_lookup: Dict[str, Type[ResourceAction]] = {
            authz.resource_action_type.__name__: authz.resource_action_type.__name__ for authz in resource_authzs
        }
        self._resource_action_lookup: Dict[str, ResourceAction] = {}
        for authz in resource_authzs:
            for action in authz.resource_action_type:
                self._resource_action_lookup[str(action)] = action
        
        self._engine = create_async_engine(**self._sqlalchemy_async_engine_kwargs)
        self._async_sessionmaker: async_sessionmaker[AsyncSession] = async_sessionmaker(
            bind=self._engine, 
            expire_on_commit=False
        )

        # For SQLite, foreign key constraints must be turned on for each connection
        if self._engine.dialect.name == "sqlite":
            @event.listens_for(self._engine.sync_engine, "connect")
            def set_sqlite_fk_pragma(dbapi_connection, connection_record):
                cursor = dbapi_connection.cursor()
                cursor.execute("PRAGMA foreign_keys=ON")
                cursor.close()
    

    def shutdown(self) -> None:
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self._engine.dispose())
    

    def setup(self) -> None:
        """Create the necessary tables for Authzee SQL storage.

        Only run this once per configuration.
        """
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self.create_tables_async())

    
    async def create_tables_async(self) -> None:
        """Create the necessary tables for Authzee SQL storage.
        """
        async with self._engine.begin() as conn:
            await conn.run_sync(Base.metadata.create_all)
 
        async with self._async_sessionmaker() as session:
            for rt_str in self._resource_type_lookup:
                session.add(ResourceTypeDB(resource_type=rt_str))
            
            for ra_str in self._resource_action_lookup:
                session.add(ResourceActionDB(resource_action=ra_str))
            
            await session.commit()


    def add_grant(self, effect: GrantEffect, grant: Grant) -> Grant:
        """Add a grant. 

        Parameters
        ----------
        effect : GrantEffect
            The effect of the grant.
        grant : Grant
            The grant.

        Returns
        -------
        Grant
            The grant that has been added with additional information for the specific backend.
        """
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(self.add_grant_async(effect=effect, grant=grant))
    

    async def add_grant_async(self, effect: GrantEffect, grant: Grant) -> Grant:
        """Add a grant. 

        Parameters
        ----------
        effect : GrantEffect
            The effect of the grant.
        grant : Grant
            The grant.

        Returns
        -------
        Grant
            The grant that has been added with additional information for the specific backend.
        """
        grant = self._check_uuid(grant=grant, generate_uuid=True)
        async with self._async_sessionmaker() as session:
            resource_action_strs = {str(action) for action in grant.resource_actions}
            result = await session.execute(
                select(ResourceActionDB).where(
                    ResourceActionDB.resource_action.in_(resource_action_strs)
                )
            )
            re_actions = set(result.scalars().fetchall())
            grant_kwargs = {
                "uuid": grant.uuid,
                "name": grant.name,
                "description": grant.description,
                "resource_type": grant.resource_type.__name__,
                "resource_actions": re_actions,
                "jmespath_expression": grant.jmespath_expression,
                "result_match": json.dumps(grant.result_match)
            }
            if effect is GrantEffect.ALLOW:
                db_grant = AllowGrantDB(**grant_kwargs)
            else:
                db_grant = DenyGrantDB(**grant_kwargs)

            session.add(db_grant)
            await session.commit()
            grant.page_id = db_grant.page_id
        
        return grant


    def delete_grant(self, effect: GrantEffect, uuid: str) -> None:
        """Delete a grant.

        Parameters
        ----------
        effect : GrantEffect
            The effect of the grant.
        uuid : str
            UUID of grant to delete.
        """
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(
            self.delete_grant_async(
                effect=effect,
                uuid=uuid
            )
        )


    async def delete_grant_async(self, effect: GrantEffect, uuid: str) -> None:
        """Delete a grant.

        Parameters
        ----------
        effect : GrantEffect
            The effect of the grant.
        uuid : str
            UUID of grant to delete.
        """
        async with self._async_sessionmaker() as session:
            if effect is GrantEffect.ALLOW:
                grant_table = AllowGrantDB
            else:
                grant_table = DenyGrantDB
            
            result = await session.execute(
                select(grant_table).where(grant_table.uuid == uuid)
            )
            db_grant = result.scalars().unique().one_or_none()
            if db_grant is None:
                raise exceptions.GrantDoesNotExistError(
                    "{} Grant with UUID: '{}' does not exist.".format(
                        effect.value,
                        uuid
                    )
                )

            await session.delete(db_grant)
            await session.commit()


    def get_grants_page(
        self,
        effect: GrantEffect,
        resource_type: Optional[Type[BaseModel]] = None,
        resource_action: Optional[ResourceAction] = None,
        page_size: Optional[int] = None,
        next_page_reference: Optional[SQLNextPageRef] = None
    ) -> GrantsPage:
        """Retrieve a page of grants matching the filters.

        If ``GrantsPage.next_page_reference`` is not ``None`` , there are more grants to retrieve.
        To get the next page, pass ``next_page_reference=GrantsPage.next_page_reference`` .

        **NOTE** - There is no guarantee of how many grants will be returned if any.

        Parameters
        ----------
        effect : GrantEffect
            The effect of the grant.
        resource_type : Optional[Type[BaseModel]], optional
            Filter by resource type.
            By default no filter is applied.
        resource_action : Optional[ResourceAction], optional
            Filter by `ResourceAction``. 
            By default no filter is applied.
        page_size : Optional[int], optional
            The suggested page size to return. 
            There is no guarantee of how much data will be returned if any.
            The default is set on the storage backend. 
        next_page_reference : Optional[SQLNextPageRef], optional
            The reference to the next page that is returned in ``GrantsPage``.
            By default this will return the 1st page.

        Returns
        -------
        GrantsPage
            The page of grants.
        """
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(
            self.get_grants_page_async(
                effect=effect,
                resource_type=resource_type,
                resource_action=resource_action,
                page_size=page_size,
                next_page_reference=next_page_reference
            )
        )
    

    async def get_grants_page_async(
        self,
        effect: GrantEffect,
        resource_type: Optional[Type[BaseModel]] = None,
        resource_action: Optional[ResourceAction] = None,
        page_size: Optional[int] = None,
        next_page_reference: Optional[SQLNextPageRef] = None
    ) -> GrantsPage:
        """Retrieve a page of grants matching the filters.

        If ``GrantsPage.next_page_reference`` is not ``None`` , there are more grants to retrieve.
        To get the next page, pass ``next_page_reference=GrantsPage.next_page_reference`` .

        **NOTE** - There is no guarantee of how many grants will be returned if any.

        Parameters
        ----------
        effect : GrantEffect
            The effect of the grant.
        resource_type : Optional[Type[BaseModel]], optional
            Filter by resource type.
            By default no filter is applied.
        resource_action : Optional[ResourceAction], optional
            Filter by `ResourceAction``. 
            By default no filter is applied.
        page_size : Optional[int], optional
            The suggested page size to return. 
            There is no guarantee of how much data will be returned if any.
            The default is set on the storage backend. 
        next_page_reference : Optional[SQLNextPageRef], optional
            The reference to the next page that is returned in ``GrantsPage``.
            By default this will return the 1st page.

        Returns
        -------
        GrantsPage
            The page of grants.
        """
        page_size = self._real_page_size(page_size=page_size)
        async with self._async_sessionmaker() as session:
            if effect is GrantEffect.ALLOW:
                grant_table = AllowGrantDB
            else:
                grant_table = DenyGrantDB

            query = select(grant_table)
            filters = []
            if resource_type is not None:
                filters.append(
                    grant_table.resource_type == resource_type.__name__
                )
            
            if resource_action is not None:
                filters.append(
                    grant_table.resource_actions.any(
                        ResourceActionDB.resource_action == str(resource_action)
                    )
                )

            if next_page_reference is not None:
                filters.append(
                    grant_table.page_id > next_page_reference.next_token
                )
            
            query = query.where(*filters)
            query = query.limit(page_size)

            result = await session.execute(query)
            db_grants = result.scalars().unique().all()
            next_page_ref = None
            if page_size is not None:
                if len(db_grants) >= page_size:
                    next_page_ref = SQLNextPageRef(next_token=db_grants[-1].page_id)
            
            grants = []
            for db_grant in db_grants:
                grants.append(
                    Grant(
                        name=db_grant.name,
                        description=db_grant.description,
                        resource_type=self._resource_type_lookup[db_grant.resource_type],
                        resource_actions={
                            self._resource_action_lookup[action.resource_action] for action in db_grant.resource_actions
                        },
                        jmespath_expression=db_grant.jmespath_expression,
                        result_match=json.loads(db_grant.result_match),
                        page_id=db_grant.page_id,
                        uuid=db_grant.uuid
                    )
                )
            
        return GrantsPage(
            grants=grants,
            next_page_reference=next_page_ref
        )







from typing import Set, Type

from pydantic import BaseModel, PrivateAttr

from authzee.resource_action import ResourceAction


class ResourceAuthz(BaseModel): 
    """Resource authorization definitions.

        - actions are the actions that can be performed on the resource.
            - action names must be unique among all resource actions
        - resource is the model for authorization of the resource
            - in many cases this can just be some sort of unique identifier
            - you can also have multiple fields though, but all fields are necessary in a resource model
        - parent_resource_authzs - list of resource authz types for parent resources
            - used in grants to authorize resources with specific parent resources
        - child_resources_authzs - list of resource authz types for child resources
            - used in grants to authorize resources with specific child resources
        

    """

    resource_type: Type[BaseModel]
    resource_action_type: Type[ResourceAction]
    parent_authz_names: Set[str]
    child_authz_names: Set[str]
    _parent_authz_types: Set[Type["ResourceAuthz"]] = PrivateAttr(default_factory=set)
    _child_authz_types: Set[Type["ResourceAuthz"]] = PrivateAttr(default_factory=set)
    _parent_resource_types: Set[Type[BaseModel]] = PrivateAttr(default_factory=set)
    _child_resource_types: Set[Type[BaseModel]] = PrivateAttr(default_factory=set)





import copy
import json
from typing import Any, AsyncIterable, Dict, Iterable, List, Optional, Set, Type, Union

import jmespath
import jmespath.exceptions
from pydantic import BaseModel

from authzee.compute.compute_backend import ComputeBackend
from authzee.jmespath_custom_functions import CustomFunctions
from authzee import exceptions
from authzee.compute import general as gc
from authzee.grant import Grant
from authzee.grant_effect import GrantEffect
from authzee.grant_iter import GrantIter
from authzee.grants_page import GrantsPage
from authzee.resource_authz import ResourceAuthz
from authzee.resource_action import ResourceAction
from authzee.storage.storage_backend import StorageBackend


class Authzee:
    """Authzee app for managing grants and verifying authorization.

    Example:

    .. code-block:: python

        from authzee import Authzee


    Parameters
    ----------
    compute_backend : ComputeBackend
        The compute backend instance.
    storage_backend : StorageBackend
        The storage backend instance.
    identity_types : Optional[Set[Type[BaseModel]]], optional
        List of identity model types to register with ``Authzee``.
        By default, none are registered.
    resource_authz_types : Optional[Set[Type[ResourceAuthz]]], optional
        List of ``ResourceAuthz`` types to register with ``Authzee``.
        By default, none are registered.
    jmespath_options : Optional[jmespath.Options], optional
        Custom JMESPath options to use for grant computations.
        See `python jmespath Options <https://github.com/jmespath/jmespath.py#options>`_ for more information.
        By default, custom functions are used from ``authzee.jmespath_custom_functions.CustomFunctions``.
    """


    def __init__(
        self, 
        compute_backend: ComputeBackend,
        storage_backend: StorageBackend,
        identity_types: Optional[Set[Type[BaseModel]]] = None,
        resource_authz_types: Optional[Set[Type[ResourceAuthz]]] = None,
        jmespath_options: Optional[jmespath.Options] = None
    ):
        self._compute_backend = compute_backend
        self._storage_backend = storage_backend
        self._both_async_enabled = (
            self._compute_backend.async_enabled 
            and self._storage_backend.async_enabled
        )
        self._identity_types: Set[Type[BaseModel]] = set()
        self._identity_type_names: Set[str] = set()
        self._resource_types: Set[Type[BaseModel]] = set()
        self._resource_type_names: Set[str] = set()
        self._authz_types: Set[Type[ResourceAuthz]] = set()
        self._authz_type_names: Set[str] = set()
        self._authzs: List[ResourceAuthz] = []
        self._resource_action_types: Set[Type[ResourceAction]] = set()
        self._resource_to_authz_lookup: Dict[Type[BaseModel], ResourceAuthz] = {}
        self._authz_name_to_authz_type_lookup: Dict[str, Type[ResourceAuthz]] = {}
        self._authz_type_to_authz_lookup: Dict[Type[ResourceAuthz], ResourceAuthz] = {}

        if identity_types is not None:
            for identity_type in identity_types:
                self.register_identity_type(identity_type=identity_type)
                
        if resource_authz_types is not None:
            for authz_type in resource_authz_types:
                self.register_resource_authz(authz_type)
        
        if jmespath_options is not None:
            self._jmespath_options = jmespath_options
        else:
            self._jmespath_options = jmespath.Options(
                custom_functions=CustomFunctions()
            )


    def initialize(self) -> None:
        """Initialize the ``Authzee`` app.

        Raises
        ------
        exceptions.InitializationError
            An error occurred while initializing the app.
        """
        for authz in self._authzs:
            for p_authz_name in authz.parent_authz_names:
                if p_authz_name not in self._authz_type_names:
                    raise exceptions.InitializationError(
                        "The parent resource '{}' in ResourceAuthz '{}' is not registered.".format(
                            p_authz_name,
                            authz.__class__.__name__
                        )
                    )

                p_authz_type = self._authz_name_to_authz_type_lookup[p_authz_name]
                authz._parent_authz_types.add(p_authz_type)
                authz._parent_resource_types.add(
                    self._authz_type_to_authz_lookup[p_authz_type].resource_type
                )

            for c_authz_name in authz.child_authz_names:
                if c_authz_name not in self._authz_type_names:
                    raise exceptions.InitializationError(
                        "The child resource '{}' in ResourceAuthz '{}' is not registered.".format(
                            c_authz_name,
                            authz.__class__.__name__
                        )
                    )

                c_authz_type = self._authz_name_to_authz_type_lookup[c_authz_name]
                authz._child_authz_types.add(c_authz_type)
                authz._child_resource_types.add(
                    self._authz_type_to_authz_lookup[c_authz_type].resource_type
                )

        # check that storage and compute are process compatible (and async?)

        self._storage_backend.initialize(
            identity_types=self._identity_types,
            resource_authzs=self._authzs
        )
        self._compute_backend.initialize(
            identity_types=self._identity_types,
            jmespath_options=self._jmespath_options,
            resource_authzs=self._authzs,
            storage_backend=self._storage_backend
        )
    
    def shutdown(self) -> None:
        """Early clean up of resources for authzee.

        If for some reason you don't want the authzee app to last the life of the program,
        you can clean up the heavier resources with this function. 
        """
        self._storage_backend.shutdown()
        self._compute_backend.shutdown()
    

    def setup(self) -> None:
        """One time setup for authzee app with the current configuration. 

        This method only has to be run once.
        """
        self._storage_backend.setup()
        self._compute_backend.setup()
    

    def teardown(self) -> None:
        """Tear down resources create for one time setup by ``setup()``.

        This may delete all storage for grants etc. 
        """
        self._storage_backend.teardown()
        self._compute_backend.teardown()



    def register_identity_type(self, identity_type: Type[BaseModel]) -> None:
        """Register an Identity model type.

        Parameters
        ----------
        identity_type : Type[BaseModel]
            Identity model type to register.

        Raises
        ------
        authzee.exceptions.IdentityRegistrationError
            Error when trying to register the Identity model type.
        """
        if identity_type in self._identity_types:
            raise exceptions.IdentityRegistrationError(
                "Identity type '{}' is already registered with Authzee".format(identity_type)
            )
        
        if identity_type.__name__ in self._identity_type_names:
            raise exceptions.IdentityRegistrationError(
                "Identity with name '{}' is already registered with Authzee".format(identity_type.__name__)
            )
        
        self._identity_types.add(identity_type)
        self._identity_type_names.add(identity_type.__name__)
    

    def register_resource_authz(self, resource_authz_type: Type[ResourceAuthz]) -> None:
        """Register a ``ResourceAuthz`` type.

        Parameters
        ----------
        resource_authz_type : Type[ResourceAuthz]
            ``ResourceAuthz`` type to register.

        Raises
        ------
        authzee.exceptions.ResourceAuthzRegistrationError
            Error when registering a ``ResourceAuthz``.
        """
        resource_authz_inst = resource_authz_type()
        if resource_authz_type in self._authz_types:
            raise exceptions.ResourceAuthzRegistrationError(
                "ResourceAuthz type '{}' is already registered with Authzee".format(resource_authz_type)
            )

        if resource_authz_type.__name__ in self._authz_type_names:
            raise exceptions.ResourceAuthzRegistrationError(
                "ResourceAuthz with name '{}' is already registered with Authzee".format(resource_authz_type.__name__)
            )

        if resource_authz_inst.resource_action_type in self._resource_action_types:
            # check the other Authz it is registered with
            registered_resource_authz = None
            for raz_inst in self._authzs:
                if raz_inst.resource_action_type == resource_authz_inst.resource_action_type:
                    registered_resource_authz = raz_inst
                    break
            
            raise exceptions.ResourceAuthzRegistrationError(
                "ResourceAction '{}' is already registered with the '{}' ResourceAuthz".format(
                    resource_authz_inst.resource_actions.__name__,
                    registered_resource_authz.__name__
                )
            )

        if resource_authz_inst.resource_type in self._resource_types:
            raise exceptions.ResourceAuthzRegistrationError(
                "Resource Model '{}' is already registered with Authzee".format(
                    resource_authz_inst.resource
                )
            )

        if resource_authz_inst.resource_type.__name__ in self._resource_type_names:
            raise exceptions.ResourceAuthzRegistrationError(
                "Resource Model with name '{}' is already registered with Authzee".format(
                    resource_authz_inst.resource
                )
            )
        
        self._resource_types.add(resource_authz_inst.resource_type)
        self._resource_type_names.add(resource_authz_inst.resource_type.__name__)
        self._resource_action_types.add(resource_authz_inst.resource_action_type)
        self._authz_types.add(resource_authz_type)
        self._authz_type_names.add(resource_authz_type.__name__)
        self._authzs.append(resource_authz_inst)
        self._resource_to_authz_lookup[resource_authz_inst.resource_type] = resource_authz_inst
        self._authz_name_to_authz_type_lookup[resource_authz_type.__name__] = resource_authz_type
        self._authz_type_to_authz_lookup[resource_authz_type] = resource_authz_inst

    
    def authorize(
        self,
        resource: BaseModel,
        resource_action: ResourceAction,
        parent_resources: List[BaseModel], 
        child_resources: List[BaseModel],
        identities: List[BaseModel],
        page_size: Optional[int] = None
    ) -> bool:
        """Authorize an entity with the given ``identities`` to perform the
        ``resource_action`` on the ``resource`` that has ``parent_resources``
        and ``child_resources``.

        Parameters
        ----------
        resource : BaseModel
            The resource model to authorize against.
        resource_action : ResourceAction
            The resource action to authorize against.
        parent_resources : List[BaseModel]
            The resource's parent resource models to authorize against.
        child_resources : List[BaseModel]
            The resource's child resource models to authorize against. 
        identities : List[BaseModel]
            The entities identities to authorize.
        page_size : Optional[int], optional
            The page size to use for the storage backend.
            The default is set on the storage backend.   

        Returns
        -------
        bool
            ``True`` if authorized, ``False`` if denied.
        
        Raises
        ------
        authzee.exceptions.InputVerificationError
            The inputs were not verified with the ``Authzee`` configuration.
        """
        self._verify_auth_args(
            resource=resource,
            resource_action=resource_action,
            parent_resources=parent_resources,
            child_resources=child_resources,
            identities=identities
        )
        jmespath_data = self._generate_jmespath_data(
            resource=resource,
            resource_action=resource_action,
            parent_resources=parent_resources,
            child_resources=child_resources,
            identities=identities
        )

        return self._compute_backend.authorize(
            resource_type=type(resource),
            resource_action=resource_action,
            jmespath_data=jmespath_data,
            page_size=page_size
        )
    

    async def authorize_async(
        self,
        resource: BaseModel,
        resource_action: ResourceAction,
        parent_resources: List[BaseModel], 
        child_resources: List[BaseModel],
        identities: List[BaseModel],
        page_size: Optional[int] = None
    ) -> bool:
        """Authorize an entity with the given ``identities`` to perform the
        ``resource_action`` on the ``resource`` that has ``parent_resources``
        and ``child_resources``.

        Parameters
        ----------
        resource : BaseModel
            The resource model to authorize against.
        resource_action : ResourceAction
            The resource action to authorize against.
        parent_resources : List[BaseModel]
            The resource's parent resource models to authorize against.
        child_resources : List[BaseModel]
            The resource's child resource models to authorize against. 
        identities : List[BaseModel]
            The entities identities to authorize.
        page_size : Optional[int], optional
            The page size to use for the storage backend.
            The default is set on the storage backend.

        Returns
        -------
        bool
            ``True`` if authorized, ``False`` if denied.

        Raises
        ------
        authzee.exceptions.AsyncNotAvailableError
            Async is not available for the storage backend.
        authzee.exceptions.InputVerificationError
            The inputs were not verified with the ``Authzee`` configuration.
        """
        if self._compute_backend.async_enabled != True:
            raise exceptions.AsyncNotAvailableError(
                "Async is not available for 'authorize' because the compute backend '{}' does not support it.".format(
                    self._compute_backend.__class__.__name__
                )
            )

        self._verify_auth_args(
            resource=resource,
            resource_action=resource_action,
            parent_resources=parent_resources,
            child_resources=child_resources,
            identities=identities
        )
        jmespath_data = self._generate_jmespath_data(
            resource=resource,
            resource_action=resource_action,
            parent_resources=parent_resources,
            child_resources=child_resources,
            identities=identities
        )

        return await self._compute_backend.authorize_async(
            resource_type=type(resource),
            resource_action=resource_action,
            jmespath_data=jmespath_data,
            page_size=page_size
        )
    


    def authorize_many(
        self,
        resources: List[BaseModel],
        resource_action: ResourceAction,
        parent_resources: List[BaseModel], 
        child_resources: List[BaseModel],
        identities: List[BaseModel],
        page_size: Optional[int] = None
    ) -> List[bool]:
        """Authorize an entity with the given ``identities`` to perform the
        ``resource_action`` on the ``resource`` s that have ``parent_resources``
        and ``child_resources``.

        Parameters
        ----------
        resources : List[BaseModel]
            The resource models to authorize against.
        resource_action : ResourceAction
            The resource action to authorize against.
        parent_resources : List[BaseModel]
            The resource's parent resource models to authorize against.
        child_resources : List[BaseModel]
            The resource's child resource models to authorize against. 
        identities : List[BaseModel]
            The entities identities to authorize.
        page_size : Optional[int], optional
            The page size to use for the storage backend.
            The default is set on the storage backend.   

        Returns
        -------
        List[bool]
            List of bools directory corresponding to ``resources``.  ``True`` if authorized, ``False`` if denied.
        
        Raises
        ------
        authzee.exceptions.InputVerificationError
            The inputs were not verified with the ``Authzee`` configuration.
        """
        self._verify_auth_many_args(
            resources=resources,
            resource_action=resource_action,
            parent_resources=parent_resources,
            child_resources=child_resources,
            identities=identities
        )
        jmespath_data = self._generate_many_jmespath_data(
            resources=resources,
            resource_action=resource_action,
            parent_resources=parent_resources,
            child_resources=child_resources,
            identities=identities
        )

        return self._compute_backend.authorize_many(
            resource_type=type(resources[0]),
            resource_action=resource_action,
            jmespath_data_entries=jmespath_data,
            page_size=page_size
        )


    async def authorize_many_async(
        self,
        resources: List[BaseModel],
        resource_action: ResourceAction,
        parent_resources: List[BaseModel], 
        child_resources: List[BaseModel],
        identities: List[BaseModel],
        page_size: Optional[int] = None
    ) -> List[bool]:
        """Authorize an entity with the given ``identities`` to perform the
        ``resource_action`` on the ``resource`` that has ``parent_resources``
        and ``child_resources``.

        Parameters
        ----------
        resources : List[BaseModel]
            The resource models to authorize against.
        resource_action : ResourceAction
            The resource action to authorize against.
        parent_resources : List[BaseModel]
            The resource's parent resource models to authorize against.
        child_resources : List[BaseModel]
            The resource's child resource models to authorize against. 
        identities : List[BaseModel]
            The entities identities to authorize.
        page_size : Optional[int], optional
            The page size to use for the storage backend.
            The default is set on the storage backend.

        Returns
        -------
        List[bool]
            List of bools directory corresponding to ``resources``.  ``True`` if authorized, ``False`` if denied.

        Raises
        ------
        authzee.exceptions.AsyncNotAvailableError
            Async is not available for the storage backend.
        authzee.exceptions.InputVerificationError
            The inputs were not verified with the ``Authzee`` configuration.
        """
        if self._compute_backend.async_enabled != True:
            raise exceptions.AsyncNotAvailableError(
                "Async is not available for 'authorize_many' because the compute backend '{}' does not support it.".format(
                    self._compute_backend.__class__.__name__
                )
            )

        self._verify_auth_many_args(
            resources=resources,
            resource_action=resource_action,
            parent_resources=parent_resources,
            child_resources=child_resources,
            identities=identities
        )
        jmespath_data = self._generate_many_jmespath_data(
            resources=resources,
            resource_action=resource_action,
            parent_resources=parent_resources,
            child_resources=child_resources,
            identities=identities
        )

        return await self._compute_backend.authorize_many_async(
            resource_type=type(resources[0]),
            resource_action=resource_action,
            jmespath_data_entries=jmespath_data,
            page_size=page_size
        )
  

    def list_grants(
        self,
        effect: GrantEffect,
        resource_type: Optional[Type[BaseModel]] = None,
        resource_action: Optional[ResourceAction] = None,
        page_size: Optional[int] = None
    ) -> Iterable[Grant]:
        """List Grants.

        Example:

        .. code-block:: python

            for grant in authzee_app.list_grants():
                print(grant.name)

        Parameters
        ----------
        effect : GrantEffect
            The effect of the grants.
        resource_type : Optional[Type[BaseModel]], optional
            Filter by resource type.
            By default no filter is applied.
        resource_action : Optional[ResourceAction], optional
            Filter by `ResourceAction``. 
            By default no filter is applied.
        page_size : Optional[int], optional
            The page size recommendation for the storage backend.
            The default is set on the storage backend. 

        Returns
        -------
        Iterable[Grant]
            An iterable for grants.
        
        Raises
        ------
        authzee.exceptions.InputVerificationError
            The inputs were not verified with the ``Authzee`` configuration.
        """
        self._verify_grant_effect(effect=effect)
        self._verify_resource_type_and_action_filter(
            resource_type=resource_type,
            resource_action=resource_action
        )

        return GrantIter(
            next_page_callable=self._storage_backend.get_grants_page,
            page_size=page_size,
            next_page_reference=None,
            effect=effect,
            resource_type=resource_type,
            resource_action=resource_action
        )


    def list_grants_async(
        self,
        effect: GrantEffect,
        resource_type: Optional[Type[BaseModel]] = None,
        resource_action: Optional[ResourceAction] = None,
        page_size: Optional[int] = None
    ) -> AsyncIterable[Grant]:
        """List Grants.

        **NOTE** - This is not a coroutine but returns an async iterator.

        Example:

        .. code-block:: python

            async for grant in authzee_app.list_grants_async():
                print(grant.name)

        Parameters
        ----------
        effect : GrantEffect
            The effect of the grants.
        resource_type : Optional[Type[BaseModel]], optional
            Filter by resource type.
            By default no filter is applied.
        resource_action : Optional[ResourceAction], optional
            Filter by `ResourceAction``. 
            By default no filter is applied.
        page_size : Optional[int], optional
            The page size recommendation for the storage backend.
            The default is set on the storage backend. 

        Returns
        -------
        AsyncIterable[Grant]
            Async iterable for grants.

        Raises
        ------
        authzee.exceptions.AsyncNotAvailableError
            Async is not available for the storage backend.
        authzee.exceptions.InputVerificationError
            The inputs were not verified with the ``Authzee`` configuration.
        """
        if self._storage_backend.async_enabled != True:
            raise exceptions.AsyncNotAvailableError(
                "Async is not available for 'list_grants' because the storage backend '{}' does not support it.".format(
                    self._storage_backend.__class__.__name__
                )
            )
        
        self._verify_grant_effect(effect=effect)
        self._verify_resource_type_and_action_filter(
            resource_type=resource_type,
            resource_action=resource_action
        )

        return GrantIter(
            next_page_callable=self._storage_backend.get_grants_page_async,
            page_size=page_size,
            next_page_reference=None,
            effect=effect,
            resource_type=resource_type,
            resource_action=resource_action
        )


    def get_grants_page(
        self,
        effect: GrantEffect,
        resource_type: Optional[Type[BaseModel]] = None,
        resource_action: Optional[ResourceAction] = None,
        page_size: Optional[int] = None,
        next_page_reference: Optional[BaseModel] = None
    ) -> GrantsPage:
        """Retrieve a page of grants matching the filters.

        If ``GrantsPage.next_page_reference`` is not ``None`` , there are more grants to retrieve.
        To get the next page, pass ``next_page_reference=GrantsPage.next_page_reference`` .

        **NOTE** - There is no guarantee of how many grants will be returned if any.

        Parameters
        ----------
        effect : GrantEffect
            The effect of the grant.
        resource_type : Optional[Type[BaseModel]], optional
            Filter by resource type.
            By default no filter is applied.
        resource_action : Optional[ResourceAction], optional
            Filter by `ResourceAction``. 
            By default no filter is applied.
        page_size : Optional[int], optional
            The suggested page size to return. 
            There is no guarantee of how much data will be returned if any.
            The default is set on the storage backend. 
        next_page_reference : Optional[BaseModel], optional
            The reference to the next page that is returned in ``GrantsPage``.
            By default this will return the 1st page.

        Returns
        -------
        GrantsPage
            The page of grants.

        Raises
        ------
        authzee.exceptions.InputVerificationError
            The inputs were not verified with the ``Authzee`` configuration.
        """
        self._verify_grant_effect(effect=effect)
        self._verify_resource_type_and_action_filter(
            resource_type=resource_type,
            resource_action=resource_action
        )

        return self._storage_backend.get_grants_page(
            effect=effect,
            resource_type=resource_type,
            resource_action=resource_action,
            page_size=page_size,
            next_page_reference=next_page_reference
        )


    async def get_grants_page_async(
        self,
        effect: GrantEffect,
        resource_type: Optional[Type[BaseModel]] = None,
        resource_action: Optional[ResourceAction] = None,
        page_size: Optional[int] = None,
        next_page_reference: Optional[BaseModel] = None
    ) -> GrantsPage:
        """Retrieve a page of grants matching the filters.

        If ``GrantsPage.next_page_reference`` is not ``None`` , there are more grants to retrieve.
        To get the next page, pass ``next_page_reference=GrantsPage.next_page_reference`` .

        **NOTE** - There is no guarantee of how many grants will be returned if any.

        Parameters
        ----------
        effect : GrantEffect
            The effect of the grant.
        resource_type : Optional[Type[BaseModel]], optional
            Filter by resource type.
            By default no filter is applied.
        resource_action : Optional[ResourceAction], optional
            Filter by `ResourceAction``. 
            By default no filter is applied.
        page_size : Optional[int], optional
            The suggested page size to return. 
            There is no guarantee of how much data will be returned if any.
            The default is set on the storage backend. 
        next_page_reference : Optional[BaseModel], optional
            The reference to the next page that is returned in ``GrantsPage``.
            By default this will return the 1st page.

        Returns
        -------
        GrantsPage
            The page of grants.
        
        Raises
        ------
        exceptions.AsyncNotAvailableError
            Async is not available for the storage backend.
        authzee.exceptions.InputVerificationError
            The inputs were not verified with the ``Authzee`` configuration.
        """
        self._verify_grant_effect(effect=effect)
        self._verify_resource_type_and_action_filter(
            resource_type=resource_type,
            resource_action=resource_action
        )

        return await self._storage_backend.get_grants_page_async(
            effect=effect,
            resource_type=resource_type,
            resource_action=resource_action,
            page_size=page_size,
            next_page_reference=next_page_reference
        )


    def list_matching_grants(
        self,
        effect: GrantEffect,
        resource: BaseModel,
        resource_action: ResourceAction,
        parent_resources: List[BaseModel], 
        child_resources: List[BaseModel],
        identities: List[BaseModel],
        page_size: Optional[int] = None
    ) -> Iterable[Grant]:
        """List matching grants.

        Parameters
        ----------
        effect : GrantEffect
            Grant effect.
        resource : BaseModel
            Resource model.
        resource_action : ResourceAction
            Resource action.
        parent_resources : List[BaseModel]
            Parent resource models.
        child_resources : List[BaseModel]
            Child resource models.
        identities : List[BaseModel]
            Identity models.
        page_size : Optional[int], optional
            The page size to use for the storage backend.
            The default is set on the storage backend.

        Returns
        -------
        Iterable[Grant]
            Iterable for matching Grants.
        
        Raises
        ------
        authzee.exceptions.InputVerificationError
            The inputs were not verified with the ``Authzee`` configuration.
        """
        self._verify_grant_effect(effect=effect)
        self._verify_auth_args(
            resource=resource,
            resource_action=resource_action,
            parent_resources=parent_resources,
            child_resources=child_resources,
            identities=identities
        )
        jmespath_data = self._generate_jmespath_data(
            resource=resource,
            resource_action=resource_action,
            parent_resources=parent_resources,
            child_resources=child_resources,
            identities=identities
        )

        return GrantIter(
            next_page_callable=self._compute_backend.get_matching_grants_page,
            page_size=page_size,
            next_page_reference=None,
            effect=effect,
            resource_type=type(resource),
            resource_action=resource_action,
            jmespath_data=jmespath_data
        )


    def list_matching_grants_async(
        self,
        effect: GrantEffect,
        resource: BaseModel,
        resource_action: ResourceAction,
        parent_resources: List[BaseModel], 
        child_resources: List[BaseModel],
        identities: List[BaseModel],
        page_size: Optional[int] = None
    ) -> AsyncIterable[Grant]:
        """List matching grants.

        **NOTE** - This is not a coroutine but returns an async iterator.

        Parameters
        ----------
        effect : GrantEffect
            Grant effect.
        resource : BaseModel
            Resource model.
        resource_action : ResourceAction
            Resource action.
        parent_resources : List[BaseModel]
            Parent resource models.
        child_resources : List[BaseModel]
            Child resource models.
        identities : List[BaseModel]
            Identity models.
        page_size : Optional[int], optional
            The page size to use for the storage backend.
            The default is set on the storage backend.


        Returns
        -------
        AsyncIterable[Grant]
            Async Iterable for matching Grants.
        
        Raises
        ------
        authzee.exceptions.AsyncNotAvailableError
            Async is not available for the storage backend.
        authzee.exceptions.InputVerificationError
            The inputs were not verified with the ``Authzee`` configuration.
        """
        if self._compute_backend.async_enabled != True:
            raise exceptions.AsyncNotAvailableError(
                "Async is not available for 'list_matching_grants' because the compute backend '{}' does not support it.".format(
                    self._compute_backend.__class__.__name__
                )
            )

        self._verify_grant_effect(effect=effect)
        self._verify_auth_args(
            resource=resource,
            resource_action=resource_action,
            parent_resources=parent_resources,
            child_resources=child_resources,
            identities=identities
        )
        jmespath_data = self._generate_jmespath_data(
            resource=resource,
            resource_action=resource_action,
            parent_resources=parent_resources,
            child_resources=child_resources,
            identities=identities
        )

        return GrantIter(
            next_page_callable=self._compute_backend.get_matching_grants_page_async,
            page_size=page_size,
            next_page_reference=None,
            effect=effect,
            resource_type=type(resource),
            resource_action=resource_action,
            jmespath_data=jmespath_data
        )


    def get_matching_grants_page(
        self,
        effect: GrantEffect,
        resource: BaseModel,
        resource_action: ResourceAction,
        parent_resources: List[BaseModel], 
        child_resources: List[BaseModel],
        identities: List[BaseModel],
        page_size: Optional[int] = None,
        next_page_reference: Optional[BaseModel] = None
    ) -> GrantsPage:
        """Retrieve a page of matching grants. 

        If ``GrantsPage.next_page_reference`` is not ``None`` , there are more grants to retrieve.
        To get the next page, pass ``next_page_reference=GrantsPage.next_page_reference`` .

        **NOTE** - There is no guarantee of how many grants will be returned if any.

        Parameters
        ----------
        effect : GrantEffect
            The effect of the grant.
        resource_type : Optional[Type[BaseModel]], optional
            Filter by resource type.
            By default no filter is applied.
        resource_action : Optional[ResourceAction], optional
            Filter by `ResourceAction``. 
            By default no filter is applied.
        jmespath_data : Dict[str, Any]
            JMESPath data that the grants will be computed with.
        page_size : Optional[int], optional
            The page size to use for the storage backend.
            This is not directly related to the returned number of grants, and can vary by compute backend.
            The default is set on the storage backend.
        next_page_reference : Optional[BaseModel], optional
            The reference to the next page that is returned in ``GrantsPage``.
            By default this will return the 1st page.

        Returns
        -------
        GrantsPage
            The page of matching grants.

        Raises
        ------
        authzee.exceptions.InputVerificationError
            The inputs were not verified with the ``Authzee`` configuration.
        """
        self._verify_grant_effect(effect=effect)
        self._verify_auth_args(
            resource=resource,
            resource_action=resource_action,
            parent_resources=parent_resources,
            child_resources=child_resources,
            identities=identities
        )
        jmespath_data = self._generate_jmespath_data(
            resource=resource,
            resource_action=resource_action,
            parent_resources=parent_resources,
            child_resources=child_resources,
            identities=identities
        )

        return self._compute_backend.get_matching_grants_page(
            effect=effect,
            resource_type=type(resource),
            resource_action=resource_action,
            jmespath_data=jmespath_data,
            page_size=page_size,
            next_page_reference=next_page_reference
        )


    async def get_matching_grants_page_async(
        self,
        effect: GrantEffect,
        resource: BaseModel,
        resource_action: ResourceAction,
        parent_resources: List[BaseModel], 
        child_resources: List[BaseModel],
        identities: List[BaseModel],
        page_size: Optional[int] = None,
        next_page_reference: Optional[BaseModel] = None
    ) -> GrantsPage:
        """Retrieve a page of matching grants. 

        If ``GrantsPage.next_page_reference`` is not ``None`` , there are more grants to retrieve.
        To get the next page, pass ``next_page_reference=GrantsPage.next_page_reference`` .

        **NOTE** - There is no guarantee of how many grants will be returned if any.

        Parameters
        ----------
        effect : GrantEffect
            The effect of the grant.
        resource_type : Optional[Type[BaseModel]], optional
            Filter by resource type.
            By default no filter is applied.
        resource_action : Optional[ResourceAction], optional
            Filter by `ResourceAction``. 
            By default no filter is applied.
        jmespath_data : Dict[str, Any]
            JMESPath data that the grants will be computed with.
        page_size : Optional[int], optional
            The page size to use for the storage backend.
            This is not directly related to the returned number of grants, and can vary by compute backend.
            The default is set on the storage backend.
        next_page_reference : Optional[BaseModel], optional
            The reference to the next page that is returned in ``GrantsPage``.
            By default this will return the 1st page.

        Returns
        -------
        GrantsPage
            The page of matching grants.

        Raises
        ------
        authzee.exceptions.AsyncNotAvailableError
            Async is not available for the storage backend.
        authzee.exceptions.InputVerificationError
            The inputs were not verified with the ``Authzee`` configuration.
        """
        if self._compute_backend.async_enabled != True:
            raise exceptions.AsyncNotAvailableError(
                "Async is not available for 'get_matching_grants_page' because the compute backend '{}' does not support it.".format(
                    self._compute_backend.__class__.__name__
                )
            )
        
        self._verify_grant_effect(effect=effect)
        self._verify_auth_args(
            resource=resource,
            resource_action=resource_action,
            parent_resources=parent_resources,
            child_resources=child_resources,
            identities=identities
        )
        jmespath_data = self._generate_jmespath_data(
            resource=resource,
            resource_action=resource_action,
            parent_resources=parent_resources,
            child_resources=child_resources,
            identities=identities
        )

        return await self._compute_backend.get_matching_grants_page_async(
            effect=effect,
            resource_type=type(resource),
            resource_action=resource_action,
            jmespath_data=jmespath_data,
            page_size=page_size,
            next_page_reference=next_page_reference
        )
    

    def add_grant(self, effect: GrantEffect, grant: Grant) -> Grant:
        """Add a grant.

        Parameters
        ----------
        effect : GrantEffect
            Effect of the grant to add.
        grant : Grant
            Grant to add.

        Returns
        -------
        Grant
            The stored grant.

        Raises
        ------
        authzee.exceptions.GrantUUIDError
            Grants that are being added should not have a UUID.
        authzee.exceptions.InputVerificationError
            The inputs were not verified with the ``Authzee`` configuration.
        """
        self._verify_grant_effect(effect=effect)
        self._verify_grant(grant=grant)
        
        return self._storage_backend.add_grant(effect=effect, grant=grant)
    

    async def add_grant_async(self, effect: GrantEffect, grant: Grant) -> Grant:
        """Add a grant.

        Parameters
        ----------
        effect : GrantEffect
            Effect of the grant to add.
        grant : Grant
            Grant to add.

        Returns
        -------
        Grant
            The stored grant.

        Raises
        ------
        authzee.exceptions.AsyncNotAvailableError
            Async is not available for the storage backend.
        authzee.exceptions.GrantUUIDError
            Grants that are being added should not have a UUID.
        authzee.exceptions.InputVerificationError
            The inputs were not verified with the ``Authzee`` configuration.
        """
        if self._compute_backend.async_enabled != True:
            raise exceptions.AsyncNotAvailableError(
                "Async is not available for 'add_grant' because the compute backend '{}' does not support it.".format(
                    self._compute_backend.__class__.__name__
                )
            )

        self._verify_grant_effect(effect=effect)
        self._verify_grant(grant=grant)

        return await self._storage_backend.add_grant_async(effect=effect, grant=grant)


    def delete_grant(self, effect: GrantEffect, uuid: str) -> None:
        """Delete a grant.

        Parameters
        ----------
        effect : GrantEffect
            Effect of the grant to delete.
        uuid : str
            UUID of grant to delete.

        Raises
        ------
        authzee.exceptions.GrantDoesNotExistError
            The given grant does not exist.
        authzee.exceptions.InputVerificationError
            The inputs were not verified with the ``Authzee`` configuration.
        """
        self._verify_grant_effect(effect=effect)
        self._storage_backend.delete_grant(effect=effect, uuid=uuid)

    
    async def delete_grant_async(self, effect: GrantEffect, uuid: str) -> None:
        """Delete a grant.

        Parameters
        ----------
        effect : GrantEffect
            Effect of the grant to delete.
        uuid : str
            UUID of grant to delete.

        Raises
        ------
        authzee.exceptions.AsyncNotAvailableError
            Async is not available for the storage backend.
        authzee.exceptions.InputVerificationError
            The inputs were not verified with the ``Authzee`` configuration.
        """
        if self._compute_backend.async_enabled != True:
            raise exceptions.AsyncNotAvailableError(
                "Async is not available for 'delete_grant' because the storage backend '{}' does not support it.".format(
                    self._compute_backend.__class__.__name__
                )
            )
        
        self._verify_grant_effect(effect=effect)
        await self._storage_backend.delete_grant_async(effect=effect, uuid=uuid)


    def grant_matches( 
        self,
        resource: BaseModel,
        resource_action: ResourceAction,
        parent_resources: List[BaseModel],
        child_resources: List[BaseModel],
        identities: List[BaseModel],
        grant: Grant
    ) -> bool:
        """Verifies a grant, the resources, and identities. Then computes if they match.

        Parameters
        ----------
        resource : BaseModel
            Resource model.
        resource_action : ResourceAction
            Resource Action.
        parent_resources : List[BaseModel]
            Parent resource models.
        child_resources : List[BaseModel]
            Child resource models.
        identities : List[BaseModel]
            Identity models.
        grant : Grant
            Grant to match against.

        Returns
        -------
        bool
            ``True`` if the grant matches, or else ``False``.
        
        Raises
        ------
        authzee.exceptions.InputVerificationError
            The inputs were not verified with the ``Authzee`` configuration.
        """
        self._verify_grant(grant=grant)
        self._verify_auth_args(
            resource=resource,
            resource_action=resource_action,
            parent_resources=parent_resources,
            child_resources=child_resources,
            identities=identities
        )
        if ( 
            type(resource) != grant.resource_type
            or resource_action not in grant.resource_actions
        ):
            return False

        jmespath_data = self._generate_jmespath_data(
            resource=resource,
            resource_action=resource_action,
            parent_resources=parent_resources,
            child_resources=child_resources,
            identities=identities
        )

        return gc.grant_matches(
            grant=grant,
            jmespath_data=jmespath_data,
            jmespath_options=self._jmespath_options
        )

    
    def _generate_jmespath_data(
        self,
        resource: BaseModel,
        resource_action: ResourceAction,
        parent_resources: List[BaseModel],
        child_resources: List[BaseModel],
        identities: List[BaseModel],
    ) -> Dict[str, Any]:
        """Generate JMESPath data.

        Parameters
        ----------
        resource : BaseModel
            Resource model.
        resource_action : ResourceAction
            Resource Action.
        parent_resources : List[BaseModel]
            Parent resource models.
        child_resources : List[BaseModel]
            Child resource models.
        identities : List[BaseModel]
            Identity models.

        Returns
        -------
        Dict[str, Any]
            The JMESPath data. 
        """
        resource_type = type(resource)
        parent_resources_by_type = {parent_type.__name__: [] for parent_type in self._resource_to_authz_lookup[resource_type]._parent_resource_types}
        for parent_resource in parent_resources:
            parent_type = type(parent_resource)
            parent_resources_by_type[parent_type.__name__].append(json.loads(parent_resource.json()))
        
        child_resources_by_type = {child_type.__name__: [] for child_type in self._resource_to_authz_lookup[resource_type]._child_resource_types}
        for child_resource in child_resources:
            child_type = type(child_resource)
            if child_type.__name__ not in child_resources_by_type:
                child_resources_by_type[child_type.__name__] = []
            
            child_resources_by_type[child_type.__name__].append(json.loads(child_resource.json()))

        identities_by_type = {identity_name: [] for identity_name in self._identity_type_names}
        for identity in identities:
            identity_type = type(identity)
            identities_by_type[identity_type.__name__].append(json.loads(identity.json()))
        
        jmespath_data = {
            "identities": identities_by_type,
            "resource": json.loads(resource.json()),
            "resource_type": type(resource).__name__,
            "resource_action": str(resource_action),
            "parent_resources": parent_resources_by_type,
            "child_resources": child_resources_by_type
        }

        return jmespath_data
    

    def _generate_many_jmespath_data(
        self,
        resources: List[BaseModel],
        resource_action: ResourceAction,
        parent_resources: List[BaseModel],
        child_resources: List[BaseModel],
        identities: List[BaseModel],
    ) -> List[Dict[str, Any]]:
        """Generate JMESPath data.

        Parameters
        ----------
        resources : List[BaseModel]
            Resource models.
        resource_action : ResourceAction
            Resource Action.
        parent_resources : List[BaseModel]
            Parent resource models.
        child_resources : List[BaseModel]
            Child resource models.
        identities : List[BaseModel]
            Identity models.

        Returns
        -------
        List[Dict[str, Any]]
            List of JMESPath data for the request. 
        """
        resource_type = type(resources[0])
        parent_resources_by_type = {parent_type.__name__: [] for parent_type in self._resource_to_authz_lookup[resource_type]._parent_resource_types}
        for parent_resource in parent_resources:
            parent_type = type(parent_resource)
            parent_resources_by_type[parent_type.__name__].append(json.loads(parent_resource.json()))
        
        child_resources_by_type = {child_type.__name__: [] for child_type in self._resource_to_authz_lookup[resource_type]._child_resource_types}
        for child_resource in child_resources:
            child_type = type(child_resource)
            if child_type.__name__ not in child_resources_by_type:
                child_resources_by_type[child_type.__name__] = []
            
            child_resources_by_type[child_type.__name__].append(json.loads(child_resource.json()))

        identities_by_type = {identity_name: [] for identity_name in self._identity_type_names}
        for identity in identities:
            identity_type = type(identity)
            identities_by_type[identity_type.__name__].append(json.loads(identity.json()))
        
        jmespath_data = {
            "identities": identities_by_type,
            "resource_type": type(resources[0]).__name__,
            "resource_action": str(resource_action),
            "parent_resources": parent_resources_by_type,
            "child_resources": child_resources_by_type
        }
        data_entries = []
        for resource in resources:
            new_jmespath_data = copy.deepcopy(jmespath_data)
            new_jmespath_data['resource'] = json.loads(resource.json())
            data_entries.append(new_jmespath_data)

        return data_entries


    def _verify_auth_args(
        self,
        resource: BaseModel,
        resource_action: ResourceAction,
        parent_resources: List[BaseModel], 
        child_resources: List[BaseModel],
        identities: List[BaseModel]
    ) -> None:
        """Verify the authorization args.

        Parameters
        ----------
        resource : BaseModel
            Resource model to verify.
        resource_action : ResourceAction
            Resource Action to verify.
        parent_resources : List[BaseModel]
            Parent resource models to verify.
        child_resources : List[BaseModel]
            Child resource models to verify.
        identities : List[BaseModel]
            Identity models to verify.

        Raises
        ------
        authzee.exceptions.InputVerificationError
            The inputs were not verified with the ``Authzee`` configuration.
        """
        resource_type = type(resource)
        self._verify_resource_type_and_action_filter(
            resource_type=resource_type,
            resource_action=resource_action
        )
        resource_authz_inst = self._resource_to_authz_lookup[resource_type]
        for parent_resource in parent_resources:
            parent_type = type(parent_resource)
            if parent_type not in self._resource_types:
                raise exceptions.InputVerificationError(
                    "Parent resource type '{}' is not a registered resource.".format(
                        parent_type.__name__
                    )
                )
            
            if type(self._resource_to_authz_lookup[parent_type]) not in resource_authz_inst._parent_authz_types:
                raise exceptions.InputVerificationError(
                    "Resource type '{}' is not a registered parent resource type of '{}'".format(
                        parent_type.__name__,
                        resource_type.__name__
                    )
                )
        
        for child_resource in child_resources:
            child_type = type(child_resource)
            if child_type not in self._resource_types:
                raise exceptions.InputVerificationError(
                    "Parent resource type '{}' is not a registered resource.".format(
                        child_type.__name__
                    )
                )
            
            if type(self._resource_to_authz_lookup[child_type]) not in resource_authz_inst._child_authz_types:
                raise exceptions.InputVerificationError(
                    "Resource type '{}' is not a registered child resource type of '{}'".format(
                        child_type.__name__,
                        resource_type.__name__
                    )
                )

        for identity in identities:
            identity_type = type(identity)
            if identity_type not in self._identity_types:
                raise exceptions.InputVerificationError(
                    "Identity type '{}' is not registered".format(
                        identity_type.__name__
                    )
                )


    def _verify_auth_many_args(
        self,
        resources: List[BaseModel],
        resource_action: ResourceAction,
        parent_resources: List[BaseModel], 
        child_resources: List[BaseModel],
        identities: List[BaseModel]
    ) -> None:
        """Verify the authorization args for many resource calls.

        Parameters
        ----------
        resources : List[BaseModel]
            Resource models to verify.
        resource_action : ResourceAction
            Resource Action to verify.
        parent_resources : List[BaseModel]
            Parent resource models to verify.
        child_resources : List[BaseModel]
            Child resource models to verify.
        identities : List[BaseModel]
            Identity models to verify.

        Raises
        ------
        authzee.exceptions.InputVerificationError
            The inputs were not verified with the ``Authzee`` configuration.
        """
        resource_type = type(resources[0])
        for resource in resources:
            if isinstance(resource, resource_type) is False:
                raise exceptions.InputVerificationError(
                    "All resources must be of the same type: {} is not of type {}".format(
                        resource, resource_type
                    )
                )
        
        self._verify_auth_args(
            resource=resources[0],
            resource_action=resource_action,
            parent_resources=parent_resources,
            child_resources=child_resources,
            identities=identities
        )


    def _verify_grant(self, grant: Grant) -> None:
        """Verify a grant with the ``Authzee`` configuration.

        Parameters
        ----------
        grant : Grant
            Grant to verify.

        Raises
        ------
        authzee.exceptions.InputVerificationError
            The inputs were not verified with the ``Authzee`` configuration.
        """
        resource_type = grant.resource_type
        if resource_type not in self._resource_types:
            raise exceptions.InputVerificationError(
                "Resource type '{}' is not a part of any registered ResourceAuthzs.".format(
                    resource_type.__name__
                )
            )
        
        if len(grant.resource_actions) < 1:
            raise exceptions.InputVerificationError("A set of at least one resource action must be given in a grant.")
        
        resource_authz_inst = self._resource_to_authz_lookup[resource_type]
        for resource_action in grant.resource_actions:
            resource_action_type = type(resource_action)
            if resource_action_type not in self._resource_action_types:
                raise exceptions.InputVerificationError(
                    "ResourceAction type '{}' is not registered.".format(
                        resource_action_type.__name__
                    )
                )

            if resource_action_type != resource_authz_inst.resource_action_type:
                raise exceptions.InputVerificationError(
                    "The '{}' resource action does not apply to the '{}' resource type.".format(
                        resource_action,
                        resource_type.__name__
                    )
                )


    def _verify_resource_type_and_action_filter(
        self, 
        resource_type: Union[Type[BaseModel], None], 
        resource_action: Union[ResourceAction, None]
    ) -> None:
        """Verify resource type and resource action filters

        Parameters
        ----------
        resource_type : Union[Type[BaseModel], None]
            Resource type model to verify, or None
        resource_action : Union[ResourceAction, None]
            Resource Action type model to verify, or None

        Raises
        ------
        authzee.exceptions.InputVerificationError
            The inputs were not verified with the ``Authzee`` configuration.
        """
        if resource_type is not None:
            if resource_type not in self._resource_types:
                raise exceptions.InputVerificationError(
                    "Resource type '{}' is not a part of any registered ResourceAuthzs.".format(
                        resource_type.__name__
                    )
                )

        if resource_action is not None:
            resource_action_type = type(resource_action)
            if resource_action_type not in self._resource_action_types:
                raise exceptions.InputVerificationError(
                    "ResourceAction type '{}' is not registered.".format(
                        resource_action_type.__name__
                    )
                )

            resource_authz_inst = self._resource_to_authz_lookup[resource_type]
            if resource_action_type != resource_authz_inst.resource_action_type:
                raise exceptions.InputVerificationError(
                    "The '{}' resource action type does not apply to the '{}' resource type.".format(
                        resource_action_type,
                        resource_type.__name__
                    )
                )



    def _verify_grant_effect(self, effect: GrantEffect) -> None:
        """Verify Grant effect type.

        Parameters
        ----------
        effect : GrantEffect
            effect to verify.

        Raises
        ------
        authzee.exceptions.InputVerificationError
            Error with grant effect type.
        """
        if type(effect) != GrantEffect:
            raise exceptions.InputVerificationError("Must use a GrantEffect, but '{}' was given.".format(effect)) 
    



from typing import Any, Dict, List, Union

from pydantic import BaseModel

from authzee.grant import Grant


class GrantsPage(BaseModel):

    grants: List[Grant]
    next_page_reference: Union[BaseModel, None]




"""Module for Authzee Exceptions
"""


class AuthzeeError(Exception):
    """Base Authzee Exception.
    """
    pass


class AsyncNotAvailableError(AuthzeeError):
    """Async is not available in a class or method
    """
    pass


class GrantDoesNotExistError(AuthzeeError):
    """The Grant Does not exist.
    """
    pass


class GrantUUIDError(AuthzeeError):
    """There was an error associated with a grant UUID
    """
    pass


class IdentityRegistrationError(AuthzeeError):
    """There was an error when registering the Identity Type.
    """
    pass


class InitializationError(AuthzeeError):
    """There was an error during initialization of the Authzee App.
    """
    pass


class InputVerificationError(AuthzeeError):
    """The given inputs could not be verified.
    """
    pass


class MethodNotImplementedError(AuthzeeError):
    """The given method is not implemented for this class.
    """

    def __init__(self, msg: str = "This method is not implemented.", *args, **kwargs):
        super().__init__(msg, *args, **kwargs)


class ResourceAuthzRegistrationError(AuthzeeError):
    """There was an error when registering the ResourceAuthz.
    """
    pass






import re
from typing import Dict, List, Union


from jmespath import exceptions, functions


class CustomFunctions(functions.Functions):
    """JMESPath custom functions.

    Along with standard `Built-in JMESPath Functions<https://jmespath.org/specification.html#built-in-functions>`_
    the following custom functions are added.

    - ``pyregex(expression: str, string: str) -> Union[str, None]`` 
        
        - Uses python's built-in ``re.fullmatch`` on the given ``string`` with the regex ``expression``
        
        - Returns the string if a full match is found or else None.

    - ``pyregex_group(expression: str, string: str, group: int) -> Union[str, None]``

        - Uses python's built-in ``re.fullmatch`` on the given set of ``strings`` with the regex ``expression``

        - Then pulls the given regex ``group`` number from the expression

        - Returns the matching regex group if found, or else None.
    
    - ``lower(string: str) -> str``
        
        - Convert string to lowercase.
    
    - ``upper(string: str) -> str``
        
        - Convert string to uppercase.

    There is also a self regulating regex cache that is added to this class.
    Because of this, **instances of this class are not thread safe** . 
    
    Parameters
    ----------
    regex_cache_size : int, optional
        Max number of compiled regex patterns to cache, by default 10000
    """

    def __init__(self, regex_cache_size: int = 10000):
        super().__init__()
        self._regex_cache_size = regex_cache_size
        self._regex_cache_count = 0
        self._regex_cache: Dict[str, re.Pattern]= {}

    
    def _get_regex(self, expression: str) -> re.Pattern:
        if expression not in self._regex_cache:
            self._regex_cache[expression] = re.compile(expression)
            self._regex_cache_count += 1
            if self._regex_cache_count > self._regex_cache_size:
                # if cache is full, remove the oldest entries
                for _ in range(self._regex_cache_count - self._regex_cache_size):
                    self._regex_cache.pop(next(iter(self._regex_cache)))
                    self._regex_cache_count -= 1
        
        return self._regex_cache[expression]


    @functions.signature(
        {"types": ["string"]}, 
        {"types": ["string"]}
    )
    def _func_pyregex(self, expression: str, string: str) -> Union[str, None]:
        if self._get_regex(expression).fullmatch(string) is not None:
            return string
        
        return None
    
    
    @functions.signature(
        {"types": ["string"]}, 
        {"types": ["string"]}, 
        {"types": ["number"]}
    )
    def _func_pyregex_group(
        self, 
        expression: str, 
        string: str, 
        group: int
    ) -> Union[str, None]:
        if type(group) != int:
            raise exceptions.JMESPathError(
                "In function pyregex_group, type of 'group' was {} but the input must translate to an integer.".format(type(group))
            )

        if group < 1:
            raise exceptions.JMESPathError("In function pyregex_group, value of 'group' was {} but the input must be greater than 0.")

        group_match = None
        re_match = self._get_regex(expression).fullmatch(string)
        if re_match is not None:
            re_groups = re_match.groups()
            if group <= len(re_groups):
                group_match = re_groups[group - 1]

        return group_match


    @functions.signature(
        {"types": ["string"]}
    )
    def _func_lower(self, string: str) -> str:
        return string.lower()


    @functions.signature(
        {"types": ["string"]}
    )
    def _func_upper(self, string: str) -> str:
        return string.upper()




from typing import Any, Optional, Set, Type, Union

from pydantic import BaseModel, validator

from authzee.resource_action import ResourceAction


class Grant(BaseModel):
    """Model for creating a grant.
    
    fill in the model
    """

    name: str
    description: str
    resource_type: Type[BaseModel] 
    resource_actions: Set[Any] 
    jmespath_expression: str
    result_match: Union[bool, dict, float, int, list, None, str] # store as json string
    page_id: Optional[int] = None
    uuid: Optional[str] = None


    @validator("resource_actions")
    def validate_actions(cls, v):
        for value in v:
            if isinstance(value, ResourceAction) != True:
                raise ValueError("'resource_actions' must come from a child class of ResourceAction")

        return v
    





from typing import Set

from sqlalchemy import Column, ForeignKey, Table
from sqlalchemy.ext.asyncio import AsyncAttrs
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, relationship


class Base(AsyncAttrs, DeclarativeBase):
    pass


class ResourceTypeDB(Base):

    __tablename__ = "resource_type"

    resource_type: Mapped[str] = mapped_column(primary_key=True)


class ResourceActionDB(Base):

    __tablename__ = "resource_action"

    resource_action: Mapped[str] = mapped_column(primary_key=True) # should be a string of the form ResourceAction.MyAction


allow_grant_action_association = Table(
    "allow_grant_action_association",
    Base.metadata,
    Column("allow_grant_page_id", ForeignKey("allow_grant.page_id"), primary_key=True),
    Column("resource_action", ForeignKey("resource_action.resource_action"), primary_key=True),
)


class AllowGrantDB(Base):

    __tablename__ = "allow_grant"

    page_id: Mapped[int] = mapped_column(primary_key=True, nullable=False)
    uuid: Mapped[str] = mapped_column(unique=True, nullable=False)
    name: Mapped[str] = mapped_column(nullable=False)
    description: Mapped[str] = mapped_column(nullable=False)
    resource_type: Mapped[str] = mapped_column(ForeignKey("resource_type.resource_type"), nullable=False)
    resource_actions: Mapped[Set[ResourceActionDB]] = relationship(
        "ResourceActionDB", 
        secondary=allow_grant_action_association, 
        lazy="joined",
        cascade=""
    )
    jmespath_expression: Mapped[str] = mapped_column(nullable=False)
    result_match: Mapped[str] = mapped_column(nullable=False)


deny_grant_action_association = Table(
    "deny_grant_action_association",
    Base.metadata,
    Column("deny_grant_page_id", ForeignKey("deny_grant.page_id"), primary_key=True),
    Column("resource_action", ForeignKey("resource_action.resource_action"), primary_key=True),
)


class DenyGrantDB(Base):

    __tablename__ = "deny_grant"

    page_id: Mapped[int] = mapped_column(primary_key=True, nullable=False)
    uuid: Mapped[str] = mapped_column(unique=True, nullable=False)
    name: Mapped[str] = mapped_column(nullable=False)
    description: Mapped[str] = mapped_column(nullable=False)
    resource_type: Mapped[str] = mapped_column(ForeignKey("resource_type.resource_type"), nullable=False)
    resource_actions: Mapped[Set[ResourceActionDB]] = relationship(
        "ResourceActionDB", 
        secondary=deny_grant_action_association, 
        lazy="joined",
        cascade=""
    )
    jmespath_expression: Mapped[str] = mapped_column(nullable=False)
    result_match: Mapped[str] = mapped_column(nullable=False)





__all__ = [
    "ComputeBackend",
    "MainProcessCompute",
    "MultiprocessCompute",
    "ThreadedCompute"
]

from authzee.compute.compute_backend import ComputeBackend

from authzee.compute.main_process_compute import MainProcessCompute
from authzee.compute.multiprocess_compute import MultiprocessCompute
from authzee.compute.threaded_compute import ThreadedCompute



from enum import Enum


class ResourceAction(Enum):
    """Base class for resource action enums

    """

    @staticmethod
    def _generate_next_value_(
        name: str, 
        start: int, 
        count: int, 
        last_values: list
    ) -> str:
        return name





from typing import Any, Dict, List, Optional, Type
import jmespath

from pydantic import BaseModel

from authzee.compute.compute_backend import ComputeBackend
from authzee.compute import general as gc
from authzee.grant_effect import GrantEffect
from authzee.grants_page import GrantsPage
from authzee.resource_action import ResourceAction
from authzee.resource_authz import ResourceAuthz
from authzee.storage.storage_backend import StorageBackend


class MainProcessCompute(ComputeBackend):

    async_enabled: bool = False
    multi_process_enabled: bool = False


    def __init__(self):
        pass


    def shutdown(self) -> None:
        """Early clean up of compute backend resources.

        NOOP
        """
        pass


    def authorize(
        self, 
        resource_type: Type[BaseModel],
        resource_action: ResourceAction,
        jmespath_data: Dict[str, Any],
        page_size: Optional[int] = None
    ) -> bool:
        done_pagination = False
        next_page_ref = None
        while done_pagination is False:
            grants_page = self._storage_backend.get_grants_page(
                effect=GrantEffect.DENY,
                resource_type=resource_type,
                resource_action=resource_action,
                next_page_reference=next_page_ref
            )
            next_page_ref = grants_page.next_page_reference
            if next_page_ref is None:
                done_pagination = True

            for grant in grants_page.grants:
                grant_match = gc.grant_matches(
                    grant=grant,
                    jmespath_data=jmespath_data,
                    jmespath_options=self._jmespath_options
                )
                if grant_match is True:
                    return False

        done_pagination = False
        next_page_ref = None
        while done_pagination is False:
            grants_page = self._storage_backend.get_grants_page(
                effect=GrantEffect.ALLOW,
                resource_type=resource_type,
                resource_action=resource_action,
                next_page_reference=next_page_ref
            )
            next_page_ref = grants_page.next_page_reference
            if next_page_ref is None:
                done_pagination = True

            for grant in grants_page.grants:
                grant_match = gc.grant_matches(
                    grant=grant,
                    jmespath_data=jmespath_data,
                    jmespath_options=self._jmespath_options
                )
                if grant_match is True:
                    return True
        
        return False


    def authorize_many(
        self, 
        resource_type: Type[BaseModel],
        resource_action: ResourceAction,
        jmespath_data_entries: List[Dict[str, Any]],
        page_size: Optional[int] = None
    ) -> List[bool]:
        results = {i: None for i in range(len(jmespath_data_entries))}
        done_pagination = False
        next_page_ref = None
        while done_pagination is False:
            grants_page = self._storage_backend.get_grants_page(
                effect=GrantEffect.DENY,
                resource_type=resource_type,
                resource_action=resource_action,
                next_page_reference=next_page_ref
            )
            next_page_ref = grants_page.next_page_reference
            if next_page_ref is None:
                done_pagination = True

            for grant in grants_page.grants:
                for i, jmespath_data in zip(results, jmespath_data_entries):
                    grant_match = gc.grant_matches(
                        grant=grant,
                        jmespath_data=jmespath_data,
                        jmespath_options=self._jmespath_options
                    )
                    if grant_match is True:
                        results[i] = False
                        values = list(results.values())
                        if None not in values:
                            return values

        done_pagination = False
        next_page_ref = None
        while done_pagination is False:
            grants_page = self._storage_backend.get_grants_page(
                effect=GrantEffect.ALLOW,
                resource_type=resource_type,
                resource_action=resource_action,
                next_page_reference=next_page_ref
            )
            next_page_ref = grants_page.next_page_reference
            if next_page_ref is None:
                done_pagination = True

            for grant in grants_page.grants:
                for i, jmespath_data in zip(results, jmespath_data_entries):
                    grant_match = gc.grant_matches(
                        grant=grant,
                        jmespath_data=jmespath_data,
                        jmespath_options=self._jmespath_options
                    )
                    if grant_match is True:
                        results[i] = True
                        values = list(results.values())
                        if None not in values:
                            return values
        
        return [val is True for val in list(results.values())]


    def get_matching_grants_page(
        self, 
        effect: GrantEffect,
        resource_type: Type[BaseModel],
        resource_action: ResourceAction,
        jmespath_data: Dict[str, Any],
        page_size: Optional[int] = None,
        next_page_reference: Optional[BaseModel] = None
    ) -> GrantsPage:
        matching_grants = []
        grants_page = self._storage_backend.get_grants_page(
            effect=effect,
            resource_type=resource_type,
            resource_action=resource_action,
            page_size=page_size,
            next_page_reference=next_page_reference
        )
        for grant in grants_page.grants:
            grant_match = gc.grant_matches(
                grant=grant,
                jmespath_data=jmespath_data,
                jmespath_options=self._jmespath_options
            )
            if grant_match == True:
                matching_grants.append(grant)
        
        return GrantsPage(
            grants=matching_grants,
            next_page_reference=grants_page.next_page_reference
        )







import copy
from typing import Dict, Iterator, List, Optional, Set, Type

from pydantic import BaseModel

from authzee import exceptions
from authzee.grant import Grant
from authzee.grant_effect import GrantEffect
from authzee.grants_page import GrantsPage
from authzee.resource_action import ResourceAction
from authzee.resource_authz import ResourceAuthz
from authzee.storage.storage_backend import StorageBackend


class MemoryStorage(StorageBackend):


    async_enabled: bool = False
    process_safe: bool = False


    def __init__(self):
        super().__init__(
            default_page_size=1000
        )


    def initialize(
        self, 
        identity_types: Set[Type[BaseModel]],
        resource_authzs: List[ResourceAuthz]
    ) -> None:
        super().initialize(identity_types, resource_authzs)
        self._allow_grants_lookup: Dict[str, Grant] = {}
        self._deny_grants_lookup: Dict[str, Grant] = {}

    
    def shutdown(self):
        pass

    
    def add_grant(self, effect: GrantEffect, grant: Grant) -> Grant:
        grant = self._check_uuid(grant=grant, generate_uuid=True)
        if effect is GrantEffect.ALLOW:
            self._allow_grants_lookup[grant.uuid] = grant
        elif effect is GrantEffect.DENY:
            self._deny_grants_lookup[grant.uuid] = grant

        return grant


    def delete_grant(self, effect: GrantEffect, uuid: str) -> None:
        if effect is GrantEffect.ALLOW:
            if uuid in self._allow_grants_lookup:
                self._allow_grants_lookup.pop(uuid)
                return
        
        if effect is GrantEffect.DENY:
            if uuid in self._deny_grants_lookup:
                self._deny_grants_lookup.pop(uuid)
                return 

        raise exceptions.GrantDoesNotExistError("{} Grant with UUID '{}' does not exist.".format(effect.value, uuid))

    
    def get_grants_page(
        self, 
        effect: GrantEffect, 
        resource_type: Optional[Type[BaseModel]] = None,
        resource_action: Optional[ResourceAction] = None,
        page_size: Optional[int] = None,
        next_page_reference: Optional[BaseModel] = None
    ) -> GrantsPage:
        if effect == GrantEffect.ALLOW:
            grants: List[Grant] = copy.deepcopy(list(self._allow_grants_lookup.values()))
        elif effect == GrantEffect.DENY:
            grants: List[Grant] = copy.deepcopy(list(self._deny_grants_lookup.values()))
        
        if resource_type is not None:
            grants = [grant for grant in grants if grant.resource_type == resource_type]
        
        if resource_action is not None:
            grants = [grant for grant in grants if resource_action in grant.resource_actions]
        
        return GrantsPage(
            grants=grants,
            next_page_reference=None
        )






__version__ = "0.1.0a2"

__all__ = [
    "Authzee",
    "Grant",
    "GrantEffect",
    "GrantsPage",
    "ResourceAction",
    "ResourceAuthz",
]

from authzee import logging_config
logging_config

from authzee.authzee import Authzee
from authzee.grant import Grant
from authzee.grant_effect import GrantEffect
from authzee.grants_page import GrantsPage
from authzee.resource_action import ResourceAction
from authzee.resource_authz import ResourceAuthz

from authzee.compute import *
from authzee.storage import *

from authzee.compute import __all__ as compute_all
from authzee.storage import __all__ as storage_all

__all__ += compute_all + storage_all






from typing import Any, Dict, List, Optional, Type

import jmespath
from pydantic import BaseModel

from authzee import exceptions
from authzee.grant_effect import GrantEffect
from authzee.grants_page import GrantsPage
from authzee.resource_action import ResourceAction
from authzee.resource_authz import ResourceAuthz
from authzee.storage.storage_backend import StorageBackend


class ComputeBackend:
    """Base class for ``Authzee`` compute backend.

    Base classes must at least implement:

        - ``initialize`` - Initialize the compute backend.
        - ``shutdown`` - Preemptively cleanup compute backend resources.
        - ``setup`` - One time setup for compute backend.
        - ``teardown`` - Remove resources created from ``setup()``.
        - ``authorize`` - Figure out if the given given identities are authorized to perform the given action on the given resource.
        - ``authorize_many`` - Figure out if the given given identities are authorized to perform the given action on the given resources.
        - ``get_matching_grants_page`` - Get a page of matching grants. 
    
    Optionally ``async`` methods may also be created for calls to compute. 

        - ``authorize_async``
        - ``authorize_many_async``
        - ``get_matching_grants_page_async``

    The sub-class must also set the class vars:

        - ``async_enabled`` - The class has all ``async`` methods available.
        - ``multi_process_enabled`` - The compute backend uses multiple processes.

    No error checking should be needed for validation of resources, resource_types etc. That should all be handled by ``Authzee``.
    """

    async_enabled: bool = False
    multi_process_enabled: bool = False


    def initialize(
        self, 
        identity_types: List[Type[BaseModel]],
        jmespath_options: jmespath.Options,
        resource_authzs: List[ResourceAuthz],
        storage_backend: StorageBackend,
    ) -> None:
        """Initialize the compute backend.

        Should only be called by the ``Authzee`` app.

        Parameters
        ----------
        identity_types : List[Type[BaseModel]]
            Identity types registered with the ``Authzee`` app.
        jmespath_options : jmespath.Options
            Custom ``jmespath.Options`` registered with the ``Authzee`` app.
            **This object should not be considered thread safe**. 
            Threads should get their own copy of this object.
            This is because custom versions of JMESPath functions are not restricted.
        resource_authzs : List[ResourceAuthz]
            ``ResourceAuthz`` s registered with the ``Authzee`` app.
        storage_backend : StorageBackend
            Storage backend registered with the ``Authzee`` app.
        """
        self._identity_types = identity_types
        self._jmespath_options = jmespath_options
        self._resource_authzs = resource_authzs
        self._storage_backend = storage_backend


    def shutdown(self) -> None:
        """Early clean up of compute backend resources.
        """
        pass 


    def setup(self) -> None:
        """One time setup for compute backend resources.
        """
        pass

    
    def teardown(self) -> None:
        """Teardown and delete the results of ``setup()`` .
        """
        pass


    def authorize(
        self, 
        resource_type: Type[BaseModel],
        resource_action: ResourceAction,
        jmespath_data: Dict[str, Any],
        page_size: Optional[int] = None,
    ) -> bool:
        """Authorize a given resource and action, with the JMESPath data against stored grants.

        First ``GrantEffect.DENY`` grants should be checked.
        If any match, then it is denied.

        Then ``GrantEffect.ALLOW`` grants are checked.
        If any match, it is allowed. If there are no matches, it is denied.

        Parameters
        ----------
        resource_type : BaseModel
            The resource type to compare grants to.
        resource_action : ResourceAction
            The resource action to compare grants to.
        jmespath_data : Dict[str, Any]
            JMESPath data that the grants will be computed with.
        page_size : Optional[int], optional
            The page size to use for the storage backend.
            The default is set on the storage backend.

        Returns
        -------
        bool
            ``True`` if allowed, ``False`` if denied.

        Raises
        ------
        authzee.exceptions.MethodNotImplementedError
            Sub-classes must implement this method.
        """
        raise exceptions.MethodNotImplementedError()


    async def authorize_async(
        self, 
        resource_type: Type[BaseModel],
        resource_action: ResourceAction,
        jmespath_data: Dict[str, Any],
        page_size: Optional[int] = None,
    ) -> bool:
        """Authorize a given resource and action, with the JMESPath data against stored grants.

        First ``GrantEffect.DENY`` grants should be checked.
        If any match, then it is denied.

        Then ``GrantEffect.ALLOW`` grants are checked.
        If any match, it is allowed. If there are no matches, it is denied.

        Parameters
        ----------
        resource_type : BaseModel
            The resource type to compare grants to.
        resource_action : ResourceAction
            The resource action to compare grants to.
        jmespath_data : Dict[str, Any]
            JMESPath data that the grants will be computed with.
        page_size : Optional[int], optional
            The page size to use for the storage backend.
            The default is set on the storage backend.

        Returns
        -------
        bool
            ``True`` if allowed, ``False`` if denied.

        Raises
        ------
        authzee.exceptions.MethodNotImplementedError
            Sub-classes *may* implement this method if ``async`` is supported.
        """
        raise exceptions.MethodNotImplementedError()
    

    def authorize_many(
        self, 
        resource_type: Type[BaseModel],
        resource_action: ResourceAction,
        jmespath_data_entries: List[Dict[str, Any]],
        page_size: Optional[int] = None,
    ) -> List[bool]:
        """Authorize a given resource and action, with the JMESPath data against stored grants.

        First ``GrantEffect.DENY`` grants should be checked.
        If any match, then it is denied.

        Then ``GrantEffect.ALLOW`` grants are checked.
        If any match, it is allowed. If there are no matches, it is denied.

        Parameters
        ----------
        resource_type : BaseModel
            The resource type to compare grants to.
        resource_action : ResourceAction
            The resource action to compare grants to.
        jmespath_data_entries : List[Dict[str, Any]]
            List of JMESPath data that the grants will be computed with.
        page_size : Optional[int], optional
            The page size to use for the storage backend.
            The default is set on the storage backend.

        Returns
        -------
        List[bool]
            List of bools directory corresponding to ``jmespath_data_entries``.  
            ``True`` if authorized, ``False`` if denied.

        Raises
        ------
        authzee.exceptions.MethodNotImplementedError
            Sub-classes must implement this method.
        """
        raise exceptions.MethodNotImplementedError()


    async def authorize_many_async(
        self, 
        resource_type: Type[BaseModel],
        resource_action: ResourceAction,
        jmespath_data_entries: List[Dict[str, Any]],
        page_size: Optional[int] = None
    ) -> List[bool]:
        """Authorize a given resource and action, with the JMESPath data against stored grants.

        First ``GrantEffect.DENY`` grants should be checked.
        If any match, then it is denied.

        Then ``GrantEffect.ALLOW`` grants are checked.
        If any match, it is allowed. If there are no matches, it is denied.

        Parameters
        ----------
        resource_type : BaseModel
            The resource type to compare grants to.
        resource_action : ResourceAction
            The resource action to compare grants to.
        jmespath_data_entries : List[Dict[str, Any]]
            List of JMESPath data that the grants will be computed with.
        page_size : Optional[int], optional
            The page size to use for the storage backend.
            The default is set on the storage backend.

        Returns
        -------
        List[bool]
            List of bools directory corresponding to ``jmespath_data_entries``.  
            ``True`` if authorized, ``False`` if denied.

        Raises
        ------
        authzee.exceptions.MethodNotImplementedError
            Sub-classes must implement this method.
        """
        raise exceptions.MethodNotImplementedError()


    def get_matching_grants_page(
        self, 
        effect: GrantEffect,
        resource_type: Type[BaseModel],
        resource_action: ResourceAction,
        jmespath_data: Dict[str, Any],
        page_size: Optional[int] = None,
        next_page_reference: Optional[BaseModel] = None
    ) -> GrantsPage:
        """Retrieve a page of matching grants. 

        If ``GrantsPage.next_page_reference`` is not ``None`` , there are more grants to retrieve.
        To get the next page, pass ``next_page_reference=GrantsPage.next_page_reference`` .

        **NOTE** - There is no guarantee of how many grants will be returned if any.

        Parameters
        ----------
        effect : GrantEffect
            The effect of the grant.
        resource_type : BaseModel
            The resource type to compare grants to.
        resource_action : ResourceAction
            The resource action to compare grants to. 
        jmespath_data : Dict[str, Any]
            JMESPath data that the grants will be computed with.
        page_size : Optional[int], optional
            The page size to use for the storage backend.
            This is not directly related to the returned number of grants, and can vary by compute backend.
            The default is set on the storage backend.
        next_page_reference : Optional[BaseModel], optional
            The reference to the next page that is returned in ``GrantsPage``.
            By default this will return the 1st page.

        Returns
        -------
        GrantsPage
            The page of matching grants.

        Raises
        ------
        authzee.exceptions.MethodNotImplementedError
            Sub-classes must implement this method.
        """
        raise exceptions.MethodNotImplementedError()


    async def get_matching_grants_page_async(
        self, 
        effect: GrantEffect,
        resource_type: Type[BaseModel],
        resource_action: ResourceAction,
        jmespath_data: Dict[str, Any],
        page_size: Optional[int] = None,
        next_page_reference: Optional[BaseModel] = None
    ) -> GrantsPage:
        """Retrieve a page of matching grants. 

        If ``GrantsPage.next_page_reference`` is not ``None`` , there are more grants to retrieve.
        To get the next page, pass ``next_page_reference=GrantsPage.next_page_reference`` .

        **NOTE** - There is no guarantee of how many grants will be returned if any.

        Parameters
        ----------
        effect : GrantEffect
            The effect of the grant.
        resource_type : BaseModel
            The resource type to compare grants to.
        resource_action : ResourceAction
            The resource action to compare grants to.
        jmespath_data : Dict[str, Any]
            JMESPath data that the grants will be computed with.
        page_size : Optional[int], optional
            The page size to use for the storage backend.
            This is not directly related to the returned number of grants, and can vary by compute backend.
            The default is set on the storage backend.
        next_page_reference : Optional[BaseModel], optional
            The reference to the next page that is returned in ``GrantsPage``.
            By default this will return the 1st page.

        Returns
        -------
        GrantsPage
            The page of matching grants.

        Raises
        ------
        authzee.exceptions.MethodNotImplementedError
            Sub-classes must implement this method.
        """
        raise exceptions.MethodNotImplementedError()








import asyncio
from concurrent.futures import ProcessPoolExecutor
import multiprocessing as mp
from multiprocessing.context import BaseContext
import os
from typing import Any, Dict, List, Optional, Type

import jmespath
from loguru import logger
from pydantic import BaseModel

from authzee.compute.compute_backend import ComputeBackend
from authzee import exceptions
from authzee.compute import general as gc
from authzee.grant import Grant
from authzee.grant_effect import GrantEffect
from authzee.grants_page import GrantsPage
from authzee.resource_action import ResourceAction
from authzee.resource_authz import ResourceAuthz
from authzee.storage.storage_backend import StorageBackend 


class MultiprocessCompute(ComputeBackend):

    async_enabled: bool = True
    multi_process_enabled: bool = False


    def __init__(
            self,
            max_workers: Optional[int] = None,
            mp_context: Optional[BaseContext] = None
        ):
        self._max_workers = max_workers
        if self._max_workers is None:
            self._max_workers = len(os.sched_getaffinity(0))
        self._mp_context = mp_context


    def initialize(
        self, 
        identity_types: List[Type[BaseModel]],
        jmespath_options: jmespath.Options,
        resource_authzs: List[ResourceAuthz],
        storage_backend: StorageBackend,
    ) -> None:
        """Initialize multiprocess backend.

        Should only be called by the ``Authzee`` app.

        Parameters
        ----------
        identity_types : List[Type[BaseModel]]
            Identity types registered with the ``Authzee`` app.
        jmespath_options : jmespath.Options
            Custom ``jmespath.Options`` registered with the ``Authzee`` app.
        resource_authzs : List[ResourceAuthz]
            ``ResourceAuthz`` s registered with the ``Authzee`` app.
        storage_backend : StorageBackend
            Storage backend registered with the ``Authzee`` app.
        """
        super().initialize(
            identity_types=identity_types,
            jmespath_options=jmespath_options,
            resource_authzs=resource_authzs,
            storage_backend=storage_backend
        )
        self._mp_manager = mp.Manager()
        self._process_pool = ProcessPoolExecutor(
            max_workers=self._max_workers, 
            mp_context=self._mp_context,
            initializer=_executor_init,
            initargs=(self._jmespath_options,)
        )
    
    def shutdown(self) -> None:
        """Early clean up of compute backend resources.

        Will shutdown the process pool without waiting for current tasks to finish.
        """
        self._process_pool.shutdown(wait=False)
        

    def authorize(
        self, 
        resource_type: Type[BaseModel],
        resource_action: ResourceAction,
        jmespath_data: Dict[str, Any],
        page_size: Optional[int] = None
    ) -> bool:
        """Authorize a given resource and action, with the JMESPath data against stored grants.

        First ``GrantEffect.DENY`` grants should be checked.
        If any match, then it is denied.

        Then ``GrantEffect.ALLOW`` grants are checked.
        If any match, it is allowed. If there are no matches, it is denied.

        Parameters
        ----------
        resource_type : BaseModel
            The resource type to compare grants to.
        resource_action : ResourceAction
            The resource action to compare grants to.
        jmespath_data : Dict[str, Any]
            JMESPath data that the grants will be computed with.
        page_size : Optional[int], optional
            The page size to use for the storage backend.
            The default is set on the storage backend.

        Returns
        -------
        bool
            ``True`` if allowed, ``False`` if denied.
        """
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(
            self.authorize_async(
                resource_type=resource_type,
                resource_action=resource_action,
                jmespath_data=jmespath_data,
                page_size=page_size
            )
        ) 


    async def authorize_async(
        self, 
        resource_type: Type[BaseModel],
        resource_action: ResourceAction,
        jmespath_data: Dict[str, Any],
        page_size: Optional[int] = None
    ) -> bool:
        """Authorize a given resource and action, with the JMESPath data against stored grants.

        First ``GrantEffect.DENY`` grants should be checked.
        If any match, then it is denied.

        Then ``GrantEffect.ALLOW`` grants are checked.
        If any match, it is allowed. If there are no matches, it is denied.

        Parameters
        ----------
        resource_type : BaseModel
            The resource type to compare grants to.
        resource_action : ResourceAction
            The resource action to compare grants to.
        jmespath_data : Dict[str, Any]
            JMESPath data that the grants will be computed with.
        page_size : Optional[int], optional
            The page size to use for the storage backend.
            The default is set on the storage backend.

        Returns
        -------
        bool
            ``True`` if allowed, ``False`` if denied.
        """ 
        loop = asyncio.get_running_loop()
        cancel_event = self._mp_manager.Event()
        done_pagination = False
        next_page_ref = None
        deny_match_event = self._mp_manager.Event()
        deny_futures: List[asyncio.Future] = []
        next_page_task = None
        grants_page = await self._storage_backend.get_grants_page_async(
            effect=GrantEffect.DENY,
            resource_type=resource_type,
            resource_action=resource_action,
            page_size=page_size,
            next_page_reference=next_page_ref
        )
        while (
            done_pagination is False
            and deny_match_event.is_set() is False
        ):
            next_page_ref = grants_page.next_page_reference
            if next_page_ref is None:
                done_pagination = True
            else:
                next_page_task = asyncio.Task(
                    self._storage_backend.get_grants_page_async(
                        effect=GrantEffect.DENY,
                        resource_type=resource_type,
                        resource_action=resource_action,
                        page_size=page_size,
                        next_page_reference=next_page_ref
                    )
                )

            deny_futures.append(
                loop.run_in_executor(
                    self._process_pool,
                    _executor_authorize,
                    grants_page, 
                    jmespath_data,
                    deny_match_event,
                    cancel_event
                )
            )
            if next_page_ref is not None:
                grants_page = await next_page_task

        done_pagination = False
        next_page_ref = None
        allow_futures = []
        allow_match_event = self._mp_manager.Event()
        next_page_task = None
        grants_page = await self._storage_backend.get_grants_page_async(
            effect=GrantEffect.ALLOW,
            resource_type=resource_type,
            resource_action=resource_action,
            page_size=page_size,
            next_page_reference=next_page_ref
        )
        while (
            done_pagination is False
            and allow_match_event.is_set() is False
            and deny_match_event.is_set() is False
        ):
            next_page_ref = grants_page.next_page_reference
            if next_page_ref is None:
                done_pagination = True
            else:
                next_page_task = asyncio.Task(
                    self._storage_backend.get_grants_page_async(
                        effect=GrantEffect.ALLOW,
                        resource_type=resource_type,
                        resource_action=resource_action,
                        page_size=page_size,
                        next_page_reference=next_page_ref
                    )
                )

            allow_futures.append(
                loop.run_in_executor(
                    self._process_pool,
                    _executor_authorize,
                    grants_page, 
                    jmespath_data,
                    allow_match_event,
                    cancel_event
                )
            )
            if next_page_ref is not None:
                grants_page = await next_page_task

        
        if deny_match_event.is_set() is True:
            cancel_event.set()
            await self._cleanup_futures(futures=deny_futures + allow_futures)

            return False
        
        if len(deny_futures) > 0:
            await asyncio.gather(*deny_futures)
            if deny_match_event.is_set() is True:
                cancel_event.set()
                await self._cleanup_futures(futures=allow_futures)

                return False
        
        if allow_match_event.is_set() is True:
            cancel_event.set()
            await self._cleanup_futures(allow_futures)
            return True

        if len(allow_futures) > 0:
            await asyncio.gather(*allow_futures)
            if allow_match_event.is_set() is True:
                return True
        
        return False


    def authorize_many(
        self, 
        resource_type: Type[BaseModel],
        resource_action: ResourceAction,
        jmespath_data_entries: List[Dict[str, Any]],
        page_size: Optional[int] = None
    ) -> List[bool]:
        """Authorize a given resource and action, with the JMESPath data against stored grants.

        First ``GrantEffect.DENY`` grants should be checked.
        If any match, then it is denied.

        Then ``GrantEffect.ALLOW`` grants are checked.
        If any match, it is allowed. If there are no matches, it is denied.

        Parameters
        ----------
        resource_type : BaseModel
            The resource type to compare grants to.
        resource_action : ResourceAction
            The resource action to compare grants to.
        jmespath_data_entries : List[Dict[str, Any]]
            List of JMESPath data that the grants will be computed with.
        page_size : Optional[int], optional
            The page size to use for the storage backend.
            The default is set on the storage backend.

        Returns
        -------
        List[bool]
            List of bools directory corresponding to ``jmespath_data_entries``.  
            ``True`` if authorized, ``False`` if denied.
        """
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(
            self.authorize_many_async(
                resource_type=resource_type,
                resource_action=resource_action,
                jmespath_data_entries=jmespath_data_entries,
                page_size=page_size
            )
        ) 


    async def authorize_many_async(
        self, 
        resource_type: Type[BaseModel],
        resource_action: ResourceAction,
        jmespath_data_entries: List[Dict[str, Any]],
        page_size: Optional[int] = None
    ) -> List[bool]:
        """Authorize a given resource and action, with the JMESPath data against stored grants.

        First ``GrantEffect.DENY`` grants should be checked.
        If any match, then it is denied.

        Then ``GrantEffect.ALLOW`` grants are checked.
        If any match, it is allowed. If there are no matches, it is denied.

        Parameters
        ----------
        resource_type : BaseModel
            The resource type to compare grants to.
        resource_action : ResourceAction
            The resource action to compare grants to.
        jmespath_data_entries : List[Dict[str, Any]]
            List of JMESPath data that the grants will be computed with.
        page_size : Optional[int], optional
            The page size to use for the storage backend.
            The default is set on the storage backend.

        Returns
        -------
        List[bool]
            List of bools directory corresponding to ``jmespath_data_entries``.  
            ``True`` if authorized, ``False`` if denied.
        """ 
        loop = asyncio.get_running_loop()
        done_pagination = False
        next_page_ref = None
        deny_futures: List[asyncio.Future] = []
        next_page_task = None
        results = {i: None for i in range(len(jmespath_data_entries))}
        grants_page = await self._storage_backend.get_grants_page_async(
            effect=GrantEffect.DENY,
            resource_type=resource_type,
            resource_action=resource_action,
            page_size=page_size,
            next_page_reference=next_page_ref
        )
        while done_pagination is False:
            next_page_ref = grants_page.next_page_reference
            if next_page_ref is None:
                done_pagination = True
            else:
                next_page_task = asyncio.Task(
                    self._storage_backend.get_grants_page_async(
                        effect=GrantEffect.DENY,
                        resource_type=resource_type,
                        resource_action=resource_action,
                        page_size=page_size,
                        next_page_reference=next_page_ref
                    )
                )

            deny_futures.append(
                loop.run_in_executor(
                    self._process_pool,
                    _executor_authorize_many,
                    grants_page, 
                    jmespath_data_entries
                )
            )
            if next_page_ref is not None:
                grants_page = await next_page_task

        done_pagination = False
        next_page_ref = None
        allow_futures = []
        next_page_task = None
        grants_page = await self._storage_backend.get_grants_page_async(
            effect=GrantEffect.ALLOW,
            resource_type=resource_type,
            resource_action=resource_action,
            page_size=page_size,
            next_page_reference=next_page_ref
        )
        while done_pagination is False:
            next_page_ref = grants_page.next_page_reference
            if next_page_ref is None:
                done_pagination = True
            else:
                next_page_task = asyncio.Task(
                    self._storage_backend.get_grants_page_async(
                        effect=GrantEffect.ALLOW,
                        resource_type=resource_type,
                        resource_action=resource_action,
                        page_size=page_size,
                        next_page_reference=next_page_ref
                    )
                )

            allow_futures.append(
                loop.run_in_executor(
                    self._process_pool,
                    _executor_authorize_many,
                    grants_page, 
                    jmespath_data_entries
                )
            )
            if next_page_ref is not None:
                grants_page = await next_page_task

        if len(deny_futures) > 0:
            deny_results: List[List[bool]] = await asyncio.gather(*deny_futures)
            for result_set in deny_results:
                for i, result in zip(results, result_set):
                    if result is True:
                        results[i] = False

        if len(allow_futures) > 0:
            allow_results: List[List[bool]] = await asyncio.gather(*allow_futures)
            for result_set in allow_results:
                for i, result in zip(results, result_set):
                    if result is True:
                        results[i] = True
        
        return [val is True for val in list(results.values())]


    def get_matching_grants_page(
        self, 
        effect: GrantEffect,
        resource_type: Type[BaseModel],
        resource_action: ResourceAction,
        jmespath_data: Dict[str, Any],
        page_size: Optional[int] = None,
        next_page_reference: Optional[BaseModel] = None
    ) -> GrantsPage:
        """Retrieve a page of matching grants. 

        If ``GrantsPage.next_page_reference`` is not ``None`` , there are more grants to retrieve.
        To get the next page, pass ``next_page_reference=GrantsPage.next_page_reference`` .

        **NOTE** - There is no guarantee of how many grants will be returned if any.

        This will send a page of grants to each worker and return the results.
        
        Parameters
        ----------
        effect : GrantEffect
            The effect of the grant.
        resource_type : BaseModel
            The resource type to compare grants to.
        resource_action : ResourceAction
            The resource action to compare grants to.
        jmespath_data : Dict[str, Any]
            JMESPath data that the grants will be computed with.
        page_size : Optional[int], optional
            The page size to use for the storage backend.
            This is not directly related to the returned number of grants, and can vary by compute backend.
            The default is set on the storage backend.
        next_page_reference : Optional[BaseModel], optional
            The reference to the next page that is returned in ``GrantsPage``.
            By default this will return the 1st page.

        Returns
        -------
        GrantsPage
            The page of matching grants.
        """
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(
            self.get_matching_grants_page_async(
                effect=effect,
                resource_type=resource_type,
                resource_action=resource_action,
                jmespath_data=jmespath_data,
                page_size=page_size,
                next_page_reference=next_page_reference
            )
        )


    async def get_matching_grants_page_async(
        self, 
        effect: GrantEffect,
        resource_type: Type[BaseModel],
        resource_action: ResourceAction,
        jmespath_data: Dict[str, Any],
        page_size: Optional[int] = None,
        next_page_reference: Optional[BaseModel] = None
    ) -> GrantsPage:
        """Retrieve a page of matching grants. 

        If ``GrantsPage.next_page_reference`` is not ``None`` , there are more grants to retrieve.
        To get the next page, pass ``next_page_reference=GrantsPage.next_page_reference`` .

        **NOTE** - There is no guarantee of how many grants will be returned if any.

        This will send a page of grants to each worker and return the results.

        Parameters
        ----------
        effect : GrantEffect
            The effect of the grant.
        resource_type : BaseModel
            The resource type to compare grants to.
        resource_action : ResourceAction
            The resource action to compare grants to.
        jmespath_data : Dict[str, Any]
            JMESPath data that the grants will be computed with.
        page_size : Optional[int], optional
            The page size to use for the storage backend.
            This is not directly related to the returned number of grants, and can vary by compute backend.
            The default is set on the storage backend.
        next_page_reference : Optional[BaseModel], optional
            The reference to the next page that is returned in ``GrantsPage``.
            By default this will return the 1st page.

        Returns
        -------
        GrantsPage
            The page of matching grants.
        """
        loop = asyncio.get_running_loop()
        futures: List[asyncio.Future] = []
        next_page_task = None
        pagination_done = False
        worker_num = 0
        grants_page =  await self._storage_backend.get_grants_page_async(
            effect=effect,
            resource_type=resource_type,
            resource_action=resource_action,
            page_size=page_size,
            next_page_reference=next_page_reference
        )
        while (
            worker_num < self._max_workers
            and pagination_done is False
        ):
            next_page_reference = grants_page.next_page_reference
            if next_page_reference is None:
                pagination_done = True

            if (
                next_page_reference is not None
                and worker_num + 1 < self._max_workers
            ):
                next_page_task = asyncio.Task(
                    self._storage_backend.get_grants_page_async(
                        effect=effect,
                        resource_type=resource_type,
                        resource_action=resource_action,
                        page_size=page_size,
                        next_page_reference=next_page_reference
                    )
                )

            futures.append(
                loop.run_in_executor(
                    self._process_pool,
                    _executor_matching_grants,
                    grants_page, 
                    jmespath_data
                )
            )
            worker_num += 1
            if (
                next_page_reference is not None
                and worker_num + 1 < self._max_workers
            ):
                grants_page = await next_page_task
        
        results = await asyncio.gather(*futures)
        
        return GrantsPage(
            grants=[grant for result in results for grant in result],
            next_page_reference=next_page_reference
        )
        

    async def _cleanup_futures(self, futures: List[asyncio.Future]) -> None:
        gather_futures: List[asyncio.Future] = []
        for future in futures:
            if future.cancel() is False:
                gather_futures.append(future)
        
        await asyncio.gather(*gather_futures)

 
def _executor_init(jmespath_options: jmespath.Options) -> None:
    global authzee_jmespath_options
    authzee_jmespath_options = jmespath_options


def _executor_authorize(
    grants_page: GrantsPage, 
    jmespath_data: Dict[str, Any], 
    match_event: mp.Event,
    cancel_event: mp.Event
) -> bool:
    global authzee_jmespath_options
    return gc.authorize_grants(
        grants_page=grants_page,
        jmespath_data=jmespath_data,
        jmespath_options=authzee_jmespath_options,
        match_event=match_event,
        cancel_event=cancel_event
    )


def _executor_authorize_many(
    grants_page: GrantsPage, 
    jmespath_data_entries: List[Dict[str, Any]]
) -> bool:
    global authzee_jmespath_options
    return gc.authorize_many_grants(
        grants_page=grants_page,
        jmespath_data_entries=jmespath_data_entries,
        jmespath_options=authzee_jmespath_options
    )


def _executor_matching_grants(
    grants_page: GrantsPage, 
    jmespath_data: Dict[str, Any]
) -> List[Grant]:
    global authzee_jmespath_options
    return gc.compute_matching_grants(
        grants_page=grants_page,
        jmespath_data=jmespath_data,
        jmespath_options=authzee_jmespath_options
    )





import multiprocessing
import threading
from typing import Any, Dict, List, Union

import jmespath
from loguru import logger

from authzee.grant import Grant
from authzee.grants_page import GrantsPage


def grant_matches(
    grant: Grant,
    jmespath_data: Dict[str, Any],
    jmespath_options: jmespath.Options
) -> bool:
    import json
    logger.debug("JMESPath Data: {}".format(json.dumps(jmespath_data, indent=4)))
    logger.debug("JMESPath Expression: {}".format(grant.jmespath_expression))
    try:
        result = jmespath.search(
            grant.jmespath_expression, 
            jmespath_data, 
            options=jmespath_options
        )
        logger.debug("JMESPath Expression Value: {}".format(result))
    except jmespath.exceptions.JMESPathError as error:
        logger.debug("JMESPath Search error: {}".format(error))
        return False

    logger.debug("JMESPath result == result_match: {}".format(result == grant.result_match))

    return result == grant.result_match


def authorize_grants(
    grants_page: GrantsPage, 
    jmespath_data: Dict[str, Any], 
    jmespath_options: jmespath.Options,
    match_event: Union[
        threading.Event,
        multiprocessing.Event
    ],
    cancel_event: Union[
        threading.Event,
        multiprocessing.Event
    ]
) -> bool:
    for grant in grants_page.grants:
        if (
            match_event.is_set() is True
            or cancel_event.is_set() is True
        ):
            return False

        grant_match = grant_matches(
            grant=grant,
            jmespath_data=jmespath_data,
            jmespath_options=jmespath_options
        )
        if grant_match is True:
            match_event.set()

            return True

    return False


def authorize_many_grants(
    grants_page: GrantsPage, 
    jmespath_data_entries: List[Dict[str, Any]], 
    jmespath_options: jmespath.Options
) -> List[Union[bool, None]]:
    results = {i: None for i in range(len(jmespath_data_entries))}
    for grant in grants_page.grants:        
        for i, jmespath_data in zip(results, jmespath_data_entries):
            grant_match = grant_matches(
                grant=grant,
                jmespath_data=jmespath_data,
                jmespath_options=jmespath_options
            )
            if grant_match is True:
                results[i] = True

    return list(results.values())


def compute_matching_grants(
    grants_page: GrantsPage, 
    jmespath_data: Dict[str, Any], 
    jmespath_options: jmespath.Options
) -> List[Grant]:
    matching_grants: List[Grant] = []
    for grant in grants_page.grants:
        grant_match = grant_matches(
            grant=grant,
            jmespath_data=jmespath_data,
            jmespath_options=jmespath_options
        )
        if grant_match is True:
            matching_grants.append(grant)

    return matching_grants
 




__all__ = [
    "StorageBackend",
    "MemoryStorage"
]

from authzee.storage.storage_backend import StorageBackend

from authzee.storage.memory_storage import MemoryStorage
try:
    from authzee.storage.sql_storage import SQLNextPageRef
    from authzee.storage.sql_storage import SQLStorage
    __all__.append("SQLNextPageRef")
    __all__.append("SQLStorage")
except ModuleNotFoundError: # pragma: no cover
    pass
